r"""
Parallel Interface to the Sage interpreter

This is an expect interface to \emph{multiple} copy of the \sage
interpreter, which can all run simultaneous calculations.  A PSage
object does not work as well as the usual Sage object, but does have
the great property that when you construct an object in a PSage you
get back a prompt immediately.  All objects constructed for that
PSage print <<currently executing code>> until code execution
completes, when they print as normal.

\note{BUG -- currently non-idle PSage subprocesses do not stop when
\sage exits.  I would very much like to fix this but don't know how.}

EXAMPLES:

We illustrate how to factor 3 integers in parallel.
First start up 3 parallel Sage interfaces::

    sage: v = [PSage() for _ in range(3)]

Next, request factorization of one random integer in each copy. ::

    sage: w = [x('factor(2^%s-1)'% randint(250,310)) for x in v]  # long time (5s on sage.math, 2011)

Print the status::

    sage: w       # long time, random output (depends on timing)
    [3 * 11 * 31^2 * 311 * 11161 * 11471 * 73471 * 715827883 * 2147483647 * 4649919401 * 18158209813151 * 5947603221397891 * 29126056043168521,
     <<currently executing code>>,
     9623 * 68492481833 * 23579543011798993222850893929565870383844167873851502677311057483194673]

Note that at the point when we printed two of the factorizations had
finished but a third one hadn't.   A few seconds later all three have
finished::

    sage: w       # long time, random output
    [3 * 11 * 31^2 * 311 * 11161 * 11471 * 73471 * 715827883 * 2147483647 * 4649919401 * 18158209813151 * 5947603221397891 * 29126056043168521,
     23^2 * 47 * 89 * 178481 * 4103188409 * 199957736328435366769577 * 44667711762797798403039426178361,
     9623 * 68492481833 * 23579543011798993222850893929565870383844167873851502677311057483194673]
"""

import os
import time

from pexpect import ExceptionPexpect

from sage.interfaces.sage0 import Sage, SageElement

number = 0


class PSage(Sage):
    def __init__(self, **kwds):
        if 'server' in kwds:
            raise NotImplementedError("PSage doesn't work on remote server yet.")
        Sage.__init__(self, **kwds)
        import sage.misc.misc
        T = sage.misc.temporary_file.tmp_dir('sage_smp')
        self.__tmp_dir = T
        self.__tmp = '%s/lock' % T
        self._unlock()
        self._unlock_code = "with open('%s', 'w') as f: f.write('__unlocked__')" % self.__tmp

        global number
        self._number = number
        number += 1

    def _repr_(self):
        """
        TESTS::

            sage: from sage.interfaces.psage import PSage
            sage: PSage()                                   # indirect doctest
            A running non-blocking (parallel) instance of Sage (number ...)
        """
        return 'A running non-blocking (parallel) instance of Sage (number %s)' % (self._number)

    def _unlock(self):
        self._locked = False
        with open(self.__tmp, 'w') as fobj:
            fobj.write('__unlocked__')

    def _lock(self):
        self._locked = True
        with open(self.__tmp, 'w') as fobj:
            fobj.write('__locked__')

    def _start(self):
        Sage._start(self)
        self.expect().timeout = 0.25
        self.expect().delaybeforesend = 0.01

    def is_locked(self) -> bool:
        try:
            with open(self.__tmp) as fobj:
                if fobj.read() != '__locked__':
                    return False
        except FileNotFoundError:
            # Directory may have already been deleted :issue:`30730`
            return False
        # looks like we are locked, but check health first
        try:
            self.expect().expect(self._prompt)
            self.expect().expect(self._prompt)
        except ExceptionPexpect:
            return False
        return True

    def __del__(self):
        """
        TESTS:

        Check that :issue:`29989` is fixed::

            sage: PSage().__del__()
        """
        try:
            files = os.listdir(self.__tmp_dir)
            for x in files:
                os.remove(os.path.join(self.__tmp_dir, x))
            os.removedirs(self.__tmp_dir)
        except OSError:
            pass

        if self._expect is not None:
            cmd = 'kill -9 %s' % self._expect.pid
            os.system(cmd)

    def eval(self, x, strip=True, **kwds):
        """
        INPUT:

        - ``x`` -- code
        - ``strip`` --ignored
        """
        if self.is_locked():
            return "<<currently executing code>>"
        if self._locked:
            self._locked = False
            # self._expect.expect('__unlocked__')
            self.expect().send('\n')
            self.expect().expect(self._prompt)
            self.expect().expect(self._prompt)
        try:
            return Sage.eval(self, x, **kwds)
        except ExceptionPexpect:
            return "<<currently executing code>>"

    def get(self, var):
        """
        Get the value of the variable var.
        """
        try:
            return self.eval('print(%s)' % var)
        except ExceptionPexpect:
            return "<<currently executing code>>"

    def set(self, var, value):
        """
        Set the variable var to the given value.
        """
        cmd = '%s=%s' % (var, value)
        self._send_nowait(cmd)
        time.sleep(0.02)

    def _send_nowait(self, x):
        if x.find('\n') != -1:
            raise ValueError("x must not have any newlines")
        # Now we want the client Python process to execute two things.
        # The first is x and the second is c.  The effect of c
        # will be to unlock the lock.
        if self.is_locked():
            return "<<currently executing code>>"
        E = self.expect()
        self._lock()
        E.write(self.preparse(x) + '\n')
        try:
            E.expect(self._prompt)
        except ExceptionPexpect:
            pass
        E.write(self._unlock_code + '\n\n')

    def _object_class(self):
        return PSageElement


class PSageElement(SageElement):
    def is_locked(self):
        return self.parent().is_locked()



r"""
Cycle enumeration

This module is meant for all functions related to cycle enumeration in graphs.

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :func:`all_cycles_iterator` | Return an iterator over the cycles of ``self``.
    :func:`all_simple_cycles` | Return an iterator over the simple cycles of ``self``.

Functions
---------
"""
# ****************************************************************************
# Copyright (C) 2025 Yuta Inoue <yutainoue888@gmail.com>
#                    David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from copy import copy


def _all_cycles_iterator_vertex(self, vertex, starting_vertices=None, simple=False,
                                rooted=False, max_length=None, trivial=False,
                                remove_acyclic_edges=True,
                                weight_function=None, by_weight=False,
                                check_weight=True, report_weight=False):
    r"""
    Return an iterator over the cycles of ``self`` starting with the given
    vertex in increasing length order. Each edge must have a positive weight.

    INPUT:

    - ``vertex`` -- the starting vertex of the cycle

    - ``starting_vertices`` -- iterable (default: ``None``); vertices from
      which the cycles must start. If ``None``, then all vertices of the
      graph can be starting points. This argument is necessary if ``rooted``
      is set to ``True``.

    - ``simple`` -- boolean (default: ``False``); if set to ``True``, then
      only simple cycles are considered. A cycle is simple if the only
      vertex occurring twice in it is the starting and ending one and no edge
      occurs twice.

    - ``rooted`` -- boolean (default: ``False``); if set to False, then
      cycles differing only by their starting vertex are considered the same
      (e.g. ``['a', 'b', 'c', 'a']`` and ``['b', 'c', 'a',
      'b']``). Otherwise, all cycles are enumerated.

    - ``max_length`` -- nonnegative integer (default: ``None``); the
      maximum length of the enumerated paths. If set to ``None``, then all
      lengths are allowed.

    - ``trivial`` -- boolean (default: ``False``); if set to ``True``, then
      the empty paths are also enumerated

    - ``remove_acyclic_edges`` -- boolean (default: ``True``); whether
      acyclic edges must be removed from the graph if ``self`` is directed.
      Used to avoid recomputing it for each vertex

    - ``weight_function`` -- function (default: ``None``); a function that
      takes as input an edge ``(u, v, l)`` and outputs its weight. If not
      ``None``, ``by_weight`` is automatically set to ``True``. If ``None``
      and ``by_weight`` is ``True``, we use the edge label ``l`` as a
      weight.

    - ``by_weight`` -- boolean (default: ``False``); if ``True``, the edges
      in the graph are weighted, otherwise all edges have weight 1

    - ``check_weight`` -- boolean (default: ``True``); whether to check that
      the ``weight_function`` outputs a number for each edge

    - ``report_weight`` -- boolean (default: ``False``); if ``False``, just
      a cycle is returned. Otherwise a tuple of cycle length and cycle is
      returned.

    OUTPUT: iterator

    EXAMPLES::

        sage: g = DiGraph({'a': ['a', 'b'], 'b': ['c'], 'c': ['d'], 'd': ['c']}, loops=True)
        sage: it = g._all_cycles_iterator_vertex('a', simple=False, max_length=None)
        sage: for i in range(5): print(next(it))
        ['a', 'a']
        ['a', 'a', 'a']
        ['a', 'a', 'a', 'a']
        ['a', 'a', 'a', 'a', 'a']
        ['a', 'a', 'a', 'a', 'a', 'a']
        sage: it = g._all_cycles_iterator_vertex('c', simple=False, max_length=None)
        sage: for i in range(5): print(next(it))
        ['c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c', 'd', 'c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c', 'd', 'c', 'd', 'c', 'd', 'c']

        sage: it = g._all_cycles_iterator_vertex('d', simple=False, max_length=None)
        sage: for i in range(5): print(next(it))
        ['d', 'c', 'd']
        ['d', 'c', 'd', 'c', 'd']
        ['d', 'c', 'd', 'c', 'd', 'c', 'd']
        ['d', 'c', 'd', 'c', 'd', 'c', 'd', 'c', 'd']
        ['d', 'c', 'd', 'c', 'd', 'c', 'd', 'c', 'd', 'c', 'd']

    It is possible to set a maximum length so that the number of cycles is
    finite::

        sage: it = g._all_cycles_iterator_vertex('d', simple=False, max_length=6)
        sage: list(it)
        [['d', 'c', 'd'], ['d', 'c', 'd', 'c', 'd'], ['d', 'c', 'd', 'c', 'd', 'c', 'd']]

    When ``simple`` is set to True, the number of cycles is finite since no vertex
    but the first one can occur more than once::

        sage: it = g._all_cycles_iterator_vertex('d', simple=True, max_length=None)
        sage: list(it)
        [['d', 'c', 'd']]

    By default, the empty cycle is not enumerated::

        sage: it = g._all_cycles_iterator_vertex('d', simple=True, trivial=True)
        sage: list(it)
        [['d'], ['d', 'c', 'd']]

    A cycle is enumerated in increasing length order for a weighted graph::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', 2), ('a', 'e', 2), ('b', 'a', 1), ('b', 'c', 1),
        ....:              ('c', 'd', 2), ('d', 'b', 1), ('e', 'a', 2)])
        sage: it = g._all_cycles_iterator_vertex('a', simple=False, max_length=None,
        ....:                                    by_weight=True, report_weight=True)
        sage: for i in range(7): print(next(it))
        (3, ['a', 'b', 'a'])
        (4, ['a', 'e', 'a'])
        (6, ['a', 'b', 'a', 'b', 'a'])
        (7, ['a', 'b', 'a', 'e', 'a'])
        (7, ['a', 'b', 'c', 'd', 'b', 'a'])
        (7, ['a', 'e', 'a', 'b', 'a'])
        (8, ['a', 'e', 'a', 'e', 'a'])

    Each edge must have a positive weight::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', -2), ('b', 'a', 1)])
        sage: next(g._all_cycles_iterator_vertex('a', simple=False, max_length=None,
        ....:                                     by_weight=True, report_weight=True))
        Traceback (most recent call last):
        ...
        ValueError: negative weight is not allowed

    The function works for an undirected graph. Specifically, each cycle is
    enumerated exactly once, meaning a cycle and its reverse are not listed separately::

        sage: g = Graph({0: [1, 2], 1: [0, 2], 2: [0, 1]})
        sage: it = g._all_cycles_iterator_vertex(0, simple=False)
        sage: for i in range(7): print(next(it))
        [0, 1, 0]
        [0, 2, 0]
        [0, 1, 2, 0]
        [0, 1, 0, 1, 0]
        [0, 1, 0, 2, 0]
        [0, 1, 2, 1, 0]
        [0, 2, 0, 2, 0]
        sage: for cycle in g._all_cycles_iterator_vertex(0, simple=True):
        ....:     print(cycle)
        [0, 1, 2, 0]
    """
    if starting_vertices is None:
        starting_vertices = [vertex]
    # First enumerate the empty cycle
    if trivial:
        if report_weight:
            yield (0, [vertex])
        else:
            yield [vertex]
    # First we remove vertices and edges that are not part of any cycle
    if self.is_directed() and remove_acyclic_edges:
        sccs = self.strongly_connected_components()
        if len(sccs) == 1:
            h = self
        else:
            d = {}
            for id, component in enumerate(sccs):
                for v in component:
                    d[v] = id
            h = copy(self)
            h.delete_edges((u, v) for u, v in h.edge_iterator(labels=False) if d[u] != d[v])
    else:
        h = self
        int_to_vertex = list(h)
        vertex_to_int = {v: i for i, v in enumerate(int_to_vertex)}

    by_weight, weight_function = self._get_weight_function(by_weight=by_weight,
                                                           weight_function=weight_function,
                                                           check_weight=check_weight)
    if by_weight:
        for e in h.edge_iterator():
            if weight_function(e) < 0:
                raise ValueError("negative weight is not allowed")

    from heapq import heappop, heappush
    heap_queue = [(0, [vertex])]
    if max_length is None:
        from sage.rings.infinity import Infinity
        max_length = Infinity
    while heap_queue:
        length, path = heappop(heap_queue)
        # Checks if a cycle has been found
        if len(path) > 1 and path[0] == path[-1]:
            report = True
            if not self.is_directed():
                if simple:
                    report = len(path) > 3 and vertex_to_int[path[1]] < vertex_to_int[path[-2]]
                else:
                    L = len(path)
                    for i in range(1, L // 2):
                        if vertex_to_int[path[i]] > vertex_to_int[path[L - i - 1]]:
                            report = False
                            break
                        if vertex_to_int[path[i]] < vertex_to_int[path[L - i - 1]]:
                            break
            if report:
                if report_weight:
                    yield (length, path)
                else:
                    yield path
        # If simple is set to True, only simple cycles are
        # allowed, Then it discards the current path
        if (not simple or path.count(path[-1]) == 1):
            for e in h.edge_iterator(vertices=[path[-1]]):
                neighbor = e[1] if e[0] == path[-1] else e[0]
                # Makes sure that the current cycle is not too long.
                # If cycles are not rooted, makes sure to keep only the
                # minimum cycle according to the lexicographic order
                if length + weight_function(e) <= max_length and \
                   (rooted or neighbor not in starting_vertices or path[0] <= neighbor):
                    heappush(heap_queue, (length + weight_function(e), path + [neighbor]))


def _all_simple_cycles_iterator_edge(self, edge, max_length=None,
                                     remove_unnecessary_edges=True,
                                     weight_function=None, by_weight=False,
                                     check_weight=True, report_weight=False):
    r"""
    Return an iterator over the **simple** cycles of ``self`` starting with the
    given edge in increasing length order. Each edge must have a positive weight.

    INPUT:

    - ``edge`` -- the starting edge of the cycle.

    - ``max_length`` -- nonnegative integer (default: ``None``); the
      maximum length of the enumerated paths. If set to ``None``, then all
      lengths are allowed.

    - ``remove_unnecessary_edges`` -- boolean (default: ``True``); whether
      unnecessary edges for enumerating simple cycles must be removed from the
      graph. If ``self`` is directed, edges not in the strongly connected
      component that contains ``edge`` are unnecessary. Otherwise, edges not in
      the 2-connected component that contains ``edge`` are unnecessary.

    - ``weight_function`` -- function (default: ``None``); a function that
      takes as input an edge ``(u, v, l)`` and outputs its weight. If not
      ``None``, ``by_weight`` is automatically set to ``True``. If ``None``
      and ``by_weight`` is ``True``, we use the edge label ``l`` as a
      weight.

    - ``by_weight`` -- boolean (default: ``False``); if ``True``, the edges
      in the graph are weighted, otherwise all edges have weight 1

    - ``check_weight`` -- boolean (default: ``True``); whether to check that
      the ``weight_function`` outputs a number for each edge

    - ``report_weight`` -- boolean (default: ``False``); if ``False``, just
      a cycle is returned. Otherwise a tuple of cycle length and cycle is
      returned.

    OUTPUT: iterator

    ALGORITHM:

    Given an edge `uv`, this algorithm extracts k-shortest `vu`-path in `G-uv`
    in increasing length order by using k-shortest path algorithm. Thus, it
    extracts only simple cycles. See
    :math:`~sage.graphs.path_enumeration.shortest_simple_paths` for more
    information.

    EXAMPLES:

        sage: g = graphs.Grid2dGraph(2, 5).to_directed()
        sage: it = g._all_simple_cycles_iterator_edge(((0, 0), (0, 1), None), report_weight=True)
        sage: for i in range(5): print(next(it))
        (2.0, [(0, 0), (0, 1), (0, 0)])
        (4.0, [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
        (6.0, [(0, 0), (0, 1), (0, 2), (1, 2), (1, 1), (1, 0), (0, 0)])
        (8.0, [(0, 0), (0, 1), (0, 2), (0, 3), (1, 3), (1, 2), (1, 1), (1, 0), (0, 0)])
        (10.0, [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (1, 4), (1, 3), (1, 2), (1, 1), (1, 0), (0, 0)])

    The function works for undirected graphs as well::

        sage: g = graphs.Grid2dGraph(2, 5)
        sage: it = g._all_simple_cycles_iterator_edge(((0, 0), (0, 1), None), report_weight=True)
        sage: for i in range(4): print(next(it))
        (4.0, [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
        (6.0, [(0, 0), (0, 1), (0, 2), (1, 2), (1, 1), (1, 0), (0, 0)])
        (8.0, [(0, 0), (0, 1), (0, 2), (0, 3), (1, 3), (1, 2), (1, 1), (1, 0), (0, 0)])
        (10.0, [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (1, 4), (1, 3), (1, 2), (1, 1), (1, 0), (0, 0)])

    Each edge must have a positive weight::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', -2), ('b', 'a', 1)])
        sage: next(g._all_simple_cycles_iterator_edge(('a', 'b', -2), max_length=None, by_weight=True))
        Traceback (most recent call last):
        ...
        ValueError: negative weight is not allowed

    TESTS:

    A graph with a loop::

        sage: g = DiGraph(loops=True)
        sage: g.add_edges([('a', 'b'), ('b', 'c'), ('c', 'a'), ('c', 'c')])
        sage: it = g._all_simple_cycles_iterator_edge(('a', 'b'), remove_unnecessary_edges=True)
        sage: next(it)
        ['a', 'b', 'c', 'a']
        sage: g = DiGraph(loops=True)
        sage: g.add_edges([('a', 'b'), ('b', 'c'), ('c', 'c')])
        sage: it = g._all_simple_cycles_iterator_edge(('c', 'c'), remove_unnecessary_edges=True)
        sage: next(it)
        ['c', 'c']

    An undirected graph with 2 biconnected components::

        sage: g = Graph()
        sage: g.add_edges([('a', 'b'), ('b', 'c'), ('c', 'a'), ('a', 'd'), ('d', 'e'), ('e', 'a')])
        sage: it = g._all_simple_cycles_iterator_edge(('a', 'b'), remove_unnecessary_edges=True)
        sage: next(it)
        ['a', 'b', 'c', 'a']
    """
    # First we remove vertices and edges that are not part of any cycle
    if remove_unnecessary_edges:
        if self.is_directed():
            components = self.strongly_connected_components_subgraphs()
        else:
            components = self.biconnected_components_subgraphs()
        h = None
        for component in components:
            if component.has_edge(edge):
                h = component
                break
        else:
            # edge connects two strongly connected components, so
            # no simple cycle starting with edge exists.
            return
    else:
        h = copy(self)
    # delete edge
    h.delete_edge(edge)

    by_weight, weight_function = self._get_weight_function(by_weight=by_weight,
                                                           weight_function=weight_function,
                                                           check_weight=check_weight)

    if by_weight:
        for e in self.edge_iterator():
            if weight_function(e) < 0:
                raise ValueError("negative weight is not allowed")

    it = h.shortest_simple_paths(source=edge[1], target=edge[0],
                                 weight_function=weight_function,
                                 by_weight=by_weight,
                                 check_weight=check_weight,
                                 report_edges=False,
                                 report_weight=True)

    edge_weight = weight_function(edge)

    if max_length is None:
        from sage.rings.infinity import Infinity
        max_length = Infinity
    for length, path in it:
        if length + edge_weight > max_length:
            break

        if report_weight:
            yield length + edge_weight, [edge[0]] + path
        else:
            yield [edge[0]] + path


def all_cycles_iterator(self, starting_vertices=None, simple=False,
                        rooted=False, max_length=None, trivial=False,
                        weight_function=None, by_weight=False,
                        check_weight=True, report_weight=False,
                        algorithm='A'):
    r"""
    Return an iterator over all the cycles of ``self`` starting with one of
    the given vertices. Each edge must have a positive weight.

    The cycles are enumerated in increasing length order. Here, a cycle
    means a closed walk `v_0e_0v_1e_1 \ldots v_{k-1}e_{k-1}v_0` where
    `v_i` are vertices and `e_i` are edges. `v_i` and `e_i`
    (`0 \leq i < k`) may not be distinct. If ``simple=True``, then cycles
    with distinct `v_i` and `e_i` (`0 \leq i < k`) are enumerated.

    INPUT:

    - ``starting_vertices`` -- iterable (default: ``None``); vertices from
      which the cycles must start. If ``None``, then all vertices of the
      graph can be starting points. This argument is necessary if ``rooted``
      is set to ``True``.

    - ``simple`` -- boolean (default: ``False``); if set to ``True``, then
      only simple cycles are considered. A cycle is simple if the only
      vertex occurring twice in it is the starting and ending one and no edge
      occurs twice.

    - ``rooted`` -- boolean (default: ``False``); if set to False, then
      cycles differing only by their starting vertex are considered the same
      (e.g. ``['a', 'b', 'c', 'a']`` and ``['b', 'c', 'a',
      'b']``). Otherwise, all cycles are enumerated.

    - ``max_length`` -- nonnegative integer (default: ``None``); the
      maximum length of the enumerated paths. If set to ``None``, then all
      lengths are allowed.

    - ``trivial`` -- boolean (default: ``False``); if set to ``True``, then
      the empty paths are also enumerated

    - ``weight_function`` -- function (default: ``None``); a function that
      takes as input an edge ``(u, v, l)`` and outputs its weight. If not
      ``None``, ``by_weight`` is automatically set to ``True``. If ``None``
      and ``by_weight`` is ``True``, we use the edge label ``l`` as a
      weight.

    - ``by_weight`` -- boolean (default: ``False``); if ``True``, the edges
      in the graph are weighted, otherwise all edges have weight 1

    - ``check_weight`` -- boolean (default: ``True``); whether to check that
      the ``weight_function`` outputs a number for each edge

    - ``report_weight`` -- boolean (default: ``False``); if ``False``, just
      a cycle is returned. Otherwise a tuple of cycle length and cycle is
      returned.

    - ``algorithm`` -- string (default: ``'A'``); the algorithm used to
      enumerate the cycles.

        - The algorithm ``'A'`` holds cycle iterators starting with each vertex,
          and output them in increasing length order.

        - The algorithm ``'B'`` holds cycle iterators starting with each edge,
          and output them in increasing length order. It depends on the k-shortest
          simple paths algorithm. Thus, it is not available if ``simple=False`` or
          ``rooted=True`` or ``starting_vertices`` is not ``None``.

    OUTPUT: iterator

    .. SEEALSO::

        - :meth:`all_simple_cycles`
        - :meth:`~sage.graphs.path_enumeration.shortest_simple_paths`

    AUTHOR:

        Alexandre Blondin Masse

    EXAMPLES::

        sage: g = DiGraph({'a': ['a', 'b'], 'b': ['c'], 'c': ['d'], 'd': ['c']}, loops=True)
        sage: it = g.all_cycles_iterator()
        sage: for _ in range(7): print(next(it))
        ['a', 'a']
        ['a', 'a', 'a']
        ['c', 'd', 'c']
        ['a', 'a', 'a', 'a']
        ['a', 'a', 'a', 'a', 'a']
        ['c', 'd', 'c', 'd', 'c']
        ['a', 'a', 'a', 'a', 'a', 'a']

    There are no cycles in the empty graph and in acyclic graphs::

        sage: g = DiGraph()
        sage: it = g.all_cycles_iterator()
        sage: list(it)
        []
        sage: g = DiGraph({0:[1]})
        sage: it = g.all_cycles_iterator()
        sage: list(it)
        []

    It is possible to restrict the starting vertices of the cycles::

        sage: g = DiGraph({'a': ['a', 'b'], 'b': ['c'], 'c': ['d'], 'd': ['c']}, loops=True)
        sage: it = g.all_cycles_iterator(starting_vertices=['b', 'c'])
        sage: for _ in range(3): print(next(it))
        ['c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c']
        ['c', 'd', 'c', 'd', 'c', 'd', 'c']

    Also, one can bound the length of the cycles::

        sage: it = g.all_cycles_iterator(max_length=3)
        sage: list(it)
        [['a', 'a'], ['a', 'a', 'a'], ['c', 'd', 'c'],
         ['a', 'a', 'a', 'a']]

    By default, cycles differing only by their starting point are not all
    enumerated, but this may be parametrized::

        sage: it = g.all_cycles_iterator(max_length=3, rooted=False)
        sage: list(it)
        [['a', 'a'], ['a', 'a', 'a'], ['c', 'd', 'c'],
         ['a', 'a', 'a', 'a']]
        sage: it = g.all_cycles_iterator(max_length=3, rooted=True)
        sage: list(it)
        [['a', 'a'], ['a', 'a', 'a'], ['c', 'd', 'c'], ['d', 'c', 'd'],
         ['a', 'a', 'a', 'a']]

    One may prefer to enumerate simple cycles, i.e. cycles such that the only
    vertex occurring twice in it is the starting and ending one (see also
    :meth:`all_simple_cycles`)::

        sage: it = g.all_cycles_iterator(simple=True)
        sage: list(it)
        [['a', 'a'], ['c', 'd', 'c']]
        sage: g = digraphs.Circuit(4)
        sage: list(g.all_cycles_iterator(simple=True))
        [[0, 1, 2, 3, 0]]

    A cycle is enumerated in increasing length order for a weighted graph::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', 2), ('a', 'e', 2), ('b', 'a', 1), ('b', 'c', 1),
        ....:              ('c', 'd', 2), ('d', 'b', 1), ('e', 'a', 2)])
        sage: it = g.all_cycles_iterator(simple=False, max_length=None,
        ....:                            by_weight=True, report_weight=True)
        sage: for i in range(9): print(next(it))
        (3, ['a', 'b', 'a'])
        (4, ['a', 'e', 'a'])
        (4, ['b', 'c', 'd', 'b'])
        (6, ['a', 'b', 'a', 'b', 'a'])
        (7, ['a', 'b', 'a', 'e', 'a'])
        (7, ['a', 'b', 'c', 'd', 'b', 'a'])
        (7, ['a', 'e', 'a', 'b', 'a'])
        (8, ['a', 'e', 'a', 'e', 'a'])
        (8, ['b', 'c', 'd', 'b', 'c', 'd', 'b'])

    Each edge must have a positive weight::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', -2), ('b', 'a', 1)])
        sage: next(g.all_cycles_iterator(simple=False, max_length=None,
        ....:                            by_weight=True, report_weight=True))
        Traceback (most recent call last):
        ...
        ValueError: negative weight is not allowed

    The algorithm ``'B'`` works for undirected graphs as well::

        sage: g = Graph({0: [1, 2], 1: [0, 2], 2: [0, 1, 3, 5], 3: [2, 4], 4: [3, 5], 5: [4, 2]})
        sage: for cycle in g.all_cycles_iterator(algorithm='B', simple=True, by_weight=True):
        ....:     print(cycle)
        [0, 1, 2, 0]
        [2, 3, 4, 5, 2]

    The algorithm ``'B'`` is unavailable when ``simple=False`` or ``rooted=True``
    or ``starting_vertices is not None``::

        sage: g = DiGraph()
        sage: g.add_edges([('a', 'b', 1), ('b', 'a', 1)])
        sage: next(g.all_cycles_iterator(algorithm='B', simple=False))
        ....:
        Traceback (most recent call last):
        ...
        ValueError: The algorithm 'B' is unavailable when simple=False.
        sage: next(g.all_cycles_iterator(algorithm='B', simple=True, rooted=True))
        Traceback (most recent call last):
        ...
        ValueError: The algorithm 'B' is unavailable when rooted=True.
        sage: next(g.all_cycles_iterator(algorithm='B', simple=True, starting_vertices=['a']))
        Traceback (most recent call last):
        ...
        ValueError: The algorithm 'B' is unavailable when starting_vertices is not None.

    The algorithm ``'A'`` works for undirected graphs as well. Specifically, each cycle is
    enumerated exactly once, meaning a cycle and its reverse are not listed separately::

        sage: g = Graph({0: [1, 2], 1: [0, 2], 2: [0, 1]})
        sage: for cycle in g.all_cycles_iterator(algorithm='A', simple=True):
        ....:     print(cycle)
        [0, 1, 2, 0]

    TESTS:

    Both algorithms support immutable graphs (:issue:`41488`)::

        sage: g = graphs.CompleteGraph(3, immutable=True)
        sage: next(g.all_cycles_iterator(algorithm='A'))
        [0, 1, 0]
        sage: next(g.all_cycles_iterator(algorithm='B', simple=True))
        [0, 1, 2, 0]
    """
    if algorithm == 'B':
        if not simple:
            raise ValueError("The algorithm 'B' is unavailable when simple=False.")
        if starting_vertices:
            raise ValueError("The algorithm 'B' is unavailable when starting_vertices is not None.")
        if rooted:
            raise ValueError("The algorithm 'B' is unavailable when rooted=True.")

    if starting_vertices is None:
        starting_vertices = self

    by_weight, weight_function = self._get_weight_function(by_weight=by_weight,
                                                           weight_function=weight_function,
                                                           check_weight=check_weight)

    if by_weight:
        for e in self.edge_iterator():
            if weight_function(e) < 0:
                raise ValueError("negative weight is not allowed")

    if algorithm == 'A':
        if self.is_directed():
            # Since a cycle is always included in a given strongly connected
            # component, we may remove edges from the graph
            sccs = self.strongly_connected_components()
            if len(sccs) == 1:
                h = self
            else:
                d = {}
                for id, component in enumerate(sccs):
                    for v in component:
                        d[v] = id
                h = copy(self)
                h.delete_edges((u, v) for u, v in h.edge_iterator(labels=False) if d[u] != d[v])
        else:
            h = self

        # We create one cycles iterator per vertex. This is necessary if we
        # want to iterate over cycles with increasing length.
        def cycle_iter(v):
            return h._all_cycles_iterator_vertex(v,
                                                 starting_vertices=starting_vertices,
                                                 simple=simple,
                                                 rooted=rooted,
                                                 max_length=max_length,
                                                 trivial=trivial,
                                                 remove_acyclic_edges=False,
                                                 weight_function=weight_function,
                                                 by_weight=by_weight,
                                                 check_weight=check_weight,
                                                 report_weight=True)

        iterators = {v: cycle_iter(v) for v in starting_vertices}
    elif algorithm == 'B':
        def simple_cycle_iter(hh, e):
            return hh._all_simple_cycles_iterator_edge(e,
                                                       max_length=max_length,
                                                       remove_unnecessary_edges=False,
                                                       weight_function=weight_function,
                                                       by_weight=by_weight,
                                                       check_weight=check_weight,
                                                       report_weight=True)
        if self.is_directed():
            def decompose(hh):
                return [hh.subgraph(c, immutable=False) for c in hh.strongly_connected_components()]
        else:
            def decompose(hh):
                return [hh.subgraph(c, immutable=False) for c in hh.biconnected_components()]

        components = decompose(self)
        iterators = dict()
        while components:
            hh = components.pop()
            if not hh.size():
                continue
            e = next(hh.edge_iterator())
            iterators[e] = simple_cycle_iter(hh, e)
            hh.delete_edge(e)
            components.extend(decompose(hh))
    else:
        raise ValueError(f"The algorithm {algorithm} is not valid. \
                            Use the algorithm 'A' or 'B'.")

    cycles = []
    for key, it in iterators.items():
        try:
            length, cycle = next(it)
            cycles.append((length, cycle, key))
        except StopIteration:
            pass
    # Since we always extract a shortest path, using a heap
    # can speed up the algorithm
    from heapq import heapify, heappop, heappush
    heapify(cycles)
    while cycles:
        # We choose the shortest available cycle
        length, shortest_cycle, key = heappop(cycles)
        if report_weight:
            yield (length, shortest_cycle)
        else:
            yield shortest_cycle
        # We update the cycle iterator to its next available cycle if it
        # exists
        try:
            length, cycle = next(iterators[key])
            heappush(cycles, (length, cycle, key))
        except StopIteration:
            pass


def all_simple_cycles(self, starting_vertices=None, rooted=False,
                      max_length=None, trivial=False,
                      weight_function=None, by_weight=False,
                      check_weight=True, report_weight=False,
                      algorithm='B'):
    r"""
    Return a list of all simple cycles of ``self``. The cycles are
    enumerated in increasing length order. Each edge must have a
    positive weight.

    INPUT:

    - ``starting_vertices`` -- iterable (default: ``None``); vertices from
      which the cycles must start. If ``None``, then all vertices of the
      graph can be starting points. This argument is necessary if ``rooted``
      is set to ``True``.

    - ``rooted`` -- boolean (default: ``False``); if set to False, then
      cycles differing only by their starting vertex are considered the same
      (e.g. ``['a', 'b', 'c', 'a']`` and ``['b', 'c', 'a',
      'b']``). Otherwise, all cycles are enumerated.

    - ``max_length`` -- nonnegative integer (default: ``None``); the
      maximum length of the enumerated paths. If set to ``None``, then all
      lengths are allowed.

    - ``trivial`` -- boolean (default: ``False``); if set to ``True``, then
      the empty paths are also enumerated

    - ``weight_function`` -- function (default: ``None``); a function that
      takes as input an edge ``(u, v, l)`` and outputs its weight. If not
      ``None``, ``by_weight`` is automatically set to ``True``. If ``None``
      and ``by_weight`` is ``True``, we use the edge label ``l`` as a
      weight.

    - ``by_weight`` -- boolean (default: ``False``); if ``True``, the edges
      in the graph are weighted, otherwise all edges have weight 1

    - ``check_weight`` -- boolean (default: ``True``); whether to check that
      the ``weight_function`` outputs a number for each edge

    - ``report_weight`` -- boolean (default: ``False``); if ``False``, just
      a cycle is returned. Otherwise a tuple of cycle length and cycle is
      returned.

    - ``algorithm`` -- string (default: ``'B'``); the algorithm used to
      enumerate the cycles.

        - The algorithm ``'A'`` holds cycle iterators starting with each vertex,
          and output them in increasing length order.

        - The algorithm ``'B'`` holds cycle iterators starting with each edge,
          and output them in increasing length order.

    OUTPUT: list

    .. NOTE::

        Although the number of simple cycles of a finite graph is always
        finite, computing all its cycles may take a very long time.

    EXAMPLES::

        sage: g = DiGraph({'a': ['a', 'b'], 'b': ['c'], 'c': ['d'], 'd': ['c']}, loops=True)
        sage: g.all_simple_cycles()
        [['a', 'a'], ['c', 'd', 'c']]

    The directed version of the Petersen graph::

        sage: g = graphs.PetersenGraph().to_directed()
        sage: g.all_simple_cycles(max_length=4)
        [[0, 1, 0], [0, 4, 0], [0, 5, 0], [1, 2, 1], [1, 6, 1], [2, 3, 2],
         [2, 7, 2], [3, 4, 3], [3, 8, 3], [4, 9, 4], [5, 7, 5], [5, 8, 5],
         [6, 8, 6], [6, 9, 6], [7, 9, 7]]
        sage: g.all_simple_cycles(max_length=6)
        [[0, 1, 0], [0, 4, 0], [0, 5, 0], [1, 2, 1], [1, 6, 1], [2, 3, 2],
         [2, 7, 2], [3, 4, 3], [3, 8, 3], [4, 9, 4], [5, 7, 5], [5, 8, 5],
         [6, 8, 6], [6, 9, 6], [7, 9, 7], [0, 1, 6, 8, 5, 0],
         [0, 1, 6, 9, 4, 0], [0, 1, 2, 7, 5, 0], [0, 1, 2, 3, 4, 0],
         [0, 4, 9, 7, 5, 0], [0, 4, 9, 6, 1, 0], [0, 4, 3, 8, 5, 0],
         [0, 4, 3, 2, 1, 0], [0, 5, 8, 6, 1, 0], [0, 5, 8, 3, 4, 0],
         [0, 5, 7, 2, 1, 0], [0, 5, 7, 9, 4, 0], [1, 2, 7, 9, 6, 1],
         [1, 2, 3, 8, 6, 1], [1, 6, 9, 7, 2, 1], [1, 6, 8, 3, 2, 1],
         [2, 3, 8, 5, 7, 2], [2, 3, 4, 9, 7, 2], [2, 7, 9, 4, 3, 2],
         [2, 7, 5, 8, 3, 2], [3, 4, 9, 6, 8, 3], [3, 8, 6, 9, 4, 3],
         [5, 7, 9, 6, 8, 5], [5, 8, 6, 9, 7, 5], [0, 1, 2, 3, 8, 5, 0],
         [0, 1, 2, 7, 9, 4, 0], [0, 1, 6, 9, 7, 5, 0],
         [0, 1, 6, 8, 3, 4, 0], [0, 4, 3, 2, 7, 5, 0],
         [0, 4, 3, 8, 6, 1, 0], [0, 4, 9, 6, 8, 5, 0],
         [0, 4, 9, 7, 2, 1, 0], [0, 5, 7, 9, 6, 1, 0],
         [0, 5, 7, 2, 3, 4, 0], [0, 5, 8, 3, 2, 1, 0],
         [0, 5, 8, 6, 9, 4, 0], [1, 2, 3, 4, 9, 6, 1],
         [1, 2, 7, 5, 8, 6, 1], [1, 6, 8, 5, 7, 2, 1],
         [1, 6, 9, 4, 3, 2, 1], [2, 3, 8, 6, 9, 7, 2],
         [2, 7, 9, 6, 8, 3, 2], [3, 4, 9, 7, 5, 8, 3],
         [3, 8, 5, 7, 9, 4, 3]]

    The complete graph (without loops) on `4` vertices::

        sage: g = graphs.CompleteGraph(4).to_directed()
        sage: g.all_simple_cycles()
        [[0, 1, 0], [0, 2, 0], [0, 3, 0], [1, 2, 1], [1, 3, 1], [2, 3, 2],
         [0, 1, 3, 0], [0, 1, 2, 0], [0, 2, 3, 0], [0, 2, 1, 0],
         [0, 3, 2, 0], [0, 3, 1, 0], [1, 2, 3, 1], [1, 3, 2, 1],
         [0, 1, 2, 3, 0], [0, 1, 3, 2, 0], [0, 2, 1, 3, 0],
         [0, 2, 3, 1, 0], [0, 3, 1, 2, 0], [0, 3, 2, 1, 0]]

    If the graph contains a large number of cycles, one can bound the length
    of the cycles, or simply restrict the possible starting vertices of the
    cycles::

        sage: g = graphs.CompleteGraph(20).to_directed()
        sage: g.all_simple_cycles(max_length=2)
        [[0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0], [0, 5, 0], [0, 6, 0], [0, 7, 0],
         [0, 8, 0], [0, 9, 0], [0, 10, 0], [0, 11, 0], [0, 12, 0], [0, 13, 0], [0, 14, 0],
         [0, 15, 0], [0, 16, 0], [0, 17, 0], [0, 18, 0], [0, 19, 0], [1, 2, 1], [1, 3, 1],
         [1, 4, 1], [1, 5, 1], [1, 6, 1], [1, 7, 1], [1, 8, 1], [1, 9, 1], [2, 3, 2],
         [2, 4, 2], [2, 5, 2], [2, 6, 2], [2, 7, 2], [2, 8, 2], [2, 9, 2], [3, 4, 3],
         [3, 5, 3], [3, 6, 3], [3, 7, 3], [3, 8, 3], [3, 9, 3], [4, 5, 4], [4, 6, 4],
         [4, 7, 4], [4, 8, 4], [4, 9, 4], [5, 6, 5], [5, 7, 5], [5, 8, 5], [5, 9, 5],
         [6, 7, 6], [6, 8, 6], [6, 9, 6], [7, 8, 7], [7, 9, 7], [8, 9, 8], [10, 1, 10],
         [10, 2, 10], [10, 3, 10], [10, 4, 10], [10, 5, 10], [10, 6, 10], [10, 7, 10],
         [10, 8, 10], [10, 9, 10], [11, 1, 11], [11, 2, 11], [11, 3, 11], [11, 4, 11],
         [11, 5, 11], [11, 6, 11], [11, 7, 11], [11, 8, 11], [11, 9, 11], [11, 10, 11],
         [12, 1, 12], [12, 2, 12], [12, 3, 12], [12, 4, 12], [12, 5, 12], [12, 6, 12],
         [12, 7, 12], [12, 8, 12], [12, 9, 12], [12, 10, 12], [12, 11, 12], [13, 1, 13],
         [13, 2, 13], [13, 3, 13], [13, 4, 13], [13, 5, 13], [13, 6, 13], [13, 7, 13],
         [13, 8, 13], [13, 9, 13], [13, 10, 13], [13, 11, 13], [13, 12, 13], [14, 1, 14],
         [14, 2, 14], [14, 3, 14], [14, 4, 14], [14, 5, 14], [14, 6, 14], [14, 7, 14],
         [14, 8, 14], [14, 9, 14], [14, 10, 14], [14, 11, 14], [14, 12, 14], [14, 13, 14],
         [15, 1, 15], [15, 2, 15], [15, 3, 15], [15, 4, 15], [15, 5, 15], [15, 6, 15],
         [15, 7, 15], [15, 8, 15], [15, 9, 15], [15, 10, 15], [15, 11, 15], [15, 12, 15],
         [15, 13, 15], [15, 14, 15], [16, 1, 16], [16, 2, 16], [16, 3, 16], [16, 4, 16],
         [16, 5, 16], [16, 6, 16], [16, 7, 16], [16, 8, 16], [16, 9, 16], [16, 10, 16],
         [16, 11, 16], [16, 12, 16], [16, 13, 16], [16, 14, 16], [16, 15, 16], [17, 1, 17],
         [17, 2, 17], [17, 3, 17], [17, 4, 17], [17, 5, 17], [17, 6, 17], [17, 7, 17],
         [17, 8, 17], [17, 9, 17], [17, 10, 17], [17, 11, 17], [17, 12, 17], [17, 13, 17],
         [17, 14, 17], [17, 15, 17], [17, 16, 17], [18, 1, 18], [18, 2, 18], [18, 3, 18],
         [18, 4, 18], [18, 5, 18], [18, 6, 18], [18, 7, 18], [18, 8, 18], [18, 9, 18],
         [18, 10, 18], [18, 11, 18], [18, 12, 18], [18, 13, 18], [18, 14, 18], [18, 15, 18],
         [18, 16, 18], [18, 17, 18], [19, 1, 19], [19, 2, 19], [19, 3, 19], [19, 4, 19],
         [19, 5, 19], [19, 6, 19], [19, 7, 19], [19, 8, 19], [19, 9, 19], [19, 10, 19],
         [19, 11, 19], [19, 12, 19], [19, 13, 19], [19, 14, 19], [19, 15, 19], [19, 16, 19],
         [19, 17, 19], [19, 18, 19]]

        sage: g = graphs.CompleteGraph(20).to_directed()
        sage: g.all_simple_cycles(max_length=2, starting_vertices=[0], algorithm='A')
        [[0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0], [0, 5, 0],
         [0, 6, 0], [0, 7, 0], [0, 8, 0], [0, 9, 0], [0, 10, 0],
         [0, 11, 0], [0, 12, 0], [0, 13, 0], [0, 14, 0], [0, 15, 0],
         [0, 16, 0], [0, 17, 0], [0, 18, 0], [0, 19, 0]]

    One may prefer to distinguish equivalent cycles having distinct starting
    vertices (compare the following examples)::

        sage: g = graphs.CompleteGraph(4).to_directed()
        sage: g.all_simple_cycles(max_length=2, rooted=False, algorithm='A')
        [[0, 1, 0], [0, 2, 0], [0, 3, 0], [1, 2, 1], [1, 3, 1], [2, 3, 2]]
        sage: g.all_simple_cycles(max_length=2, rooted=True, algorithm='A')
        [[0, 1, 0], [0, 2, 0], [0, 3, 0], [1, 0, 1], [1, 2, 1], [1, 3, 1],
         [2, 0, 2], [2, 1, 2], [2, 3, 2], [3, 0, 3], [3, 1, 3], [3, 2, 3]]

    A cycle is enumerated in increasing length order for a weighted graph::

        sage: cycles = g.all_simple_cycles(weight_function=lambda e:e[0]+e[1],
        ....:                              by_weight=True, report_weight=True)
        sage: cycles
        [(2.0, [0, 1, 0]), (4.0, [0, 2, 0]), (6.0, [0, 1, 2, 0]), (6.0, [0, 2, 1, 0]),
         (6.0, [0, 3, 0]), (6.0, [1, 2, 1]), (8.0, [0, 1, 3, 0]), (8.0, [0, 3, 1, 0]),
         (8.0, [1, 3, 1]), (10.0, [0, 2, 3, 0]), (10.0, [0, 3, 2, 0]), (10.0, [2, 3, 2]),
         (12.0, [0, 1, 3, 2, 0]), (12.0, [0, 1, 2, 3, 0]), (12.0, [0, 2, 3, 1, 0]),
         (12.0, [0, 2, 1, 3, 0]), (12.0, [0, 3, 2, 1, 0]), (12.0, [0, 3, 1, 2, 0]),
         (12.0, [1, 2, 3, 1]), (12.0, [1, 3, 2, 1])]

    The algorithm ``'B'`` can be used::

        sage: cycles_B = g.all_simple_cycles(weight_function=lambda e:e[0]+e[1], by_weight=True,
        ....:                                report_weight=True, algorithm='B')
        sage: cycles_B
        [(2.0, [0, 1, 0]), (4.0, [0, 2, 0]), (6.0, [0, 1, 2, 0]), (6.0, [0, 2, 1, 0]),
         (6.0, [0, 3, 0]), (6.0, [1, 2, 1]), (8.0, [0, 1, 3, 0]), (8.0, [0, 3, 1, 0]),
         (8.0, [1, 3, 1]), (10.0, [0, 2, 3, 0]), (10.0, [0, 3, 2, 0]), (10.0, [2, 3, 2]),
         (12.0, [0, 1, 3, 2, 0]), (12.0, [0, 1, 2, 3, 0]), (12.0, [0, 2, 3, 1, 0]),
         (12.0, [0, 2, 1, 3, 0]), (12.0, [0, 3, 2, 1, 0]), (12.0, [0, 3, 1, 2, 0]),
         (12.0, [1, 2, 3, 1]), (12.0, [1, 3, 2, 1])]
        sage: cycles.sort() == cycles_B.sort()
        True

    The algorithm ``'A'`` is available for undirected graphs. Specifically, each cycle is
    enumerated exactly once, meaning a cycle and its reverse are not listed separately::

        sage: g = Graph({0: [1, 2], 1: [0, 2], 2: [0, 1]})
        sage: g.all_simple_cycles(algorithm='A')
        [[0, 1, 2, 0]]
    """
    return list(self.all_cycles_iterator(starting_vertices=starting_vertices,
                                         simple=True, rooted=rooted,
                                         max_length=max_length, trivial=trivial,
                                         weight_function=weight_function,
                                         by_weight=by_weight,
                                         check_weight=check_weight,
                                         report_weight=report_weight,
                                         algorithm=algorithm))



# distutils: libraries = glpk z gmp

#*****************************************************************************
#       Copyright (C) 2010 Nathann Cohen <nathann.cohen@gmail.com>
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.libs.glpk.types cimport *

cdef extern from "glpk.h":
    void glp_init_iocp(glp_iocp *)
    void glp_init_smcp(glp_smcp *)
    glp_prob * glp_create_prob()
    void glp_set_prob_name(glp_prob *, char *)
    const char* glp_get_prob_name(glp_prob *)
    void glp_set_obj_dir(glp_prob *, int)
    void glp_add_rows(glp_prob *, int)
    void glp_add_cols(glp_prob *, int)
    void glp_del_rows(glp_prob *, int, int *)
    void glp_set_row_name(glp_prob *, int, char *)
    void glp_set_col_name(glp_prob *, int, char *)
    void glp_set_row_bnds(glp_prob *, int, int, double, double)
    void glp_set_col_bnds(glp_prob *, int, int, double, double)
    void glp_set_obj_coef(glp_prob *, int, double)
    void glp_set_row_stat(glp_prob *, int, int)
    void glp_set_col_stat(glp_prob *, int, int)
    int glp_warm_up(glp_prob *)
    void glp_load_matrix(glp_prob *, int, int *, int *, double *)
    int glp_simplex(glp_prob *, glp_smcp *)
    int glp_exact(glp_prob *, glp_smcp *)
    int glp_intopt(glp_prob *, glp_iocp *)
    int lpx_intopt(glp_prob *)
    void glp_std_basis(glp_prob *)
    void glp_delete_prob(glp_prob *)
    double glp_get_col_prim(glp_prob *, int)
    double glp_get_obj_val(glp_prob *)
    double glp_get_col_dual(glp_prob *, int)
    double glp_get_row_prim(glp_prob *, int)
    double glp_get_row_dual(glp_prob *, int)
    int glp_get_col_stat(glp_prob *, int)
    int glp_get_row_stat(glp_prob *, int)
    int glp_print_ranges(glp_prob *lp, int,int, int, char *fname)
    int glp_get_num_rows(glp_prob *)
    int glp_get_num_cols(glp_prob *)
    double glp_mip_col_val(glp_prob *, int)
    double glp_mip_obj_val(glp_prob *)
    void glp_set_col_kind(glp_prob *, int, int)
    int glp_write_mps(glp_prob *lp, int fmt, void *parm, char *fname)
    int glp_write_lp(glp_prob *lp, void *parm, char *fname)
    void glp_set_prob_name(glp_prob *lp, char *name)
    void glp_set_obj_name(glp_prob *lp, char *name)
    void glp_set_row_name(glp_prob *lp, int i, char *name)
    void glp_set_col_name(glp_prob *lp, int i, char *name)

    double glp_get_row_ub(glp_prob *lp, int i)
    double glp_get_row_lb(glp_prob *lp, int i)

    double glp_get_col_ub(glp_prob *lp, int i)
    double glp_get_col_lb(glp_prob *lp, int i)

    void glp_set_col_ub(glp_prob *lp, int i, double value)
    void glp_set_col_lb(glp_prob *lp, int i, double value)

    int glp_eval_tab_row(glp_prob *lp, int k, int ind[], double val[])
    int glp_eval_tab_col(glp_prob *lp, int k, int ind[], double val[])

    const char* glp_get_row_name(glp_prob *lp, int i)
    const char* glp_get_col_name(glp_prob *lp, int i)

    void glp_create_index(glp_prob *lp)

    int glp_get_prim_stat(glp_prob *lp)
    int glp_get_status(glp_prob *lp)
    int glp_mip_status(glp_prob *lp)
    int glp_set_mat_row(glp_prob *lp, int, int, int *, double * )
    int glp_set_mat_col(glp_prob *lp, int, int, int *, double * )
    int glp_get_mat_row(glp_prob *lp, int, int *, double * )
    double glp_get_row_ub(glp_prob *lp, int)
    double glp_get_row_lb(glp_prob *lp, int)
    int glp_get_col_kind(glp_prob *lp, int)
    double glp_get_obj_coef(glp_prob *lp, int)
    int glp_get_obj_dir(glp_prob *lp)
    void glp_copy_prob(glp_prob *dst, glp_prob *src, int names)
    double glp_ios_mip_gap(glp_tree *T)
    int glp_ios_best_node(glp_tree *tree)
    double glp_ios_node_bound(glp_tree *T, int p)

    int glp_bf_exists(glp_prob *lp)



# sage.doctest: needs sage.combinat sage.modules
"""
Symmetric functions defined by orthogonality and triangularity

One characterization of Schur functions is that they are upper
triangularly related to the monomial symmetric functions and
orthogonal with respect to the Hall scalar product. We can use the
class SymmetricFunctionAlgebra_orthotriang to obtain the Schur
functions from this definition.

::

    sage: from sage.combinat.sf.sfa import zee
    sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
    sage: Sym = SymmetricFunctions(QQ)
    sage: m = Sym.m()
    sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
    sage: s([2,1])^2
    s[2, 2, 1, 1] + s[2, 2, 2] + s[3, 1, 1, 1] + 2*s[3, 2, 1] + s[3, 3] + s[4, 1, 1] + s[4, 2]

::

    sage: s2 = SymmetricFunctions(QQ).s()
    sage: s2([2,1])^2
    s[2, 2, 1, 1] + s[2, 2, 2] + s[3, 1, 1, 1] + 2*s[3, 2, 1] + s[3, 3] + s[4, 1, 1] + s[4, 2]
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.homset import Hom
from sage.categories.morphism import SetMorphism

from . import sfa


class SymmetricFunctionAlgebra_orthotriang(sfa.SymmetricFunctionAlgebra_generic):

    class Element(sfa.SymmetricFunctionAlgebra_generic.Element):
        pass

    @staticmethod
    def __classcall__(cls, Sym, base, scalar, prefix, basis_name, leading_coeff=None):
        """
        Normalize the arguments.

        TESTS::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: B1 = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur')
            sage: B2 = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur', None)
            sage: B1 is B2
            True
        """
        return super().__classcall__(cls, Sym, base, scalar, prefix, basis_name, leading_coeff)

    def __init__(self, Sym, base, scalar, prefix, basis_name, leading_coeff):
        r"""
        Initialization of the symmetric function algebra defined via orthotriangular rules.

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``Sym`` -- ring of symmetric functions
        - ``base`` -- an instance of a basis of the ring of symmetric functions
          (e.g. the Schur functions)
        - ``scalar`` -- a function ``zee`` on partitions. The function
          ``zee`` determines the scalar product on the power sum basis
          with normalization `\langle p_{\mu}, p_{\mu} \rangle =
          \mathrm{zee}(\mu)`.
        - ``prefix`` -- the prefix used to display the basis
        - ``basis_name`` -- the name used for the basis

        .. NOTE::

            The base ring is required to be a `\QQ`-algebra for this
            method to be usable, since the scalar product is defined by
            its values on the power sum basis.

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur'); s
            Symmetric Functions over Rational Field in the Schur basis

        TESTS::

            sage: TestSuite(s).run(elements=[s[1,1]+2*s[2], s[1]+3*s[1,1]])
            sage: TestSuite(s).run(skip=["_test_associativity", "_test_prod"])  # long time (7s on sage.math, 2011)

        Note: ``s.an_element()`` is of degree 4; so we skip
        ``_test_associativity`` and ``_test_prod`` which involve
        (currently?) expensive calculations up to degree 12.
        """
        self._sym = Sym
        self._sf_base = base
        self._scalar = scalar
        self._leading_coeff = leading_coeff
        sfa.SymmetricFunctionAlgebra_generic.__init__(self, Sym, prefix=prefix, basis_name=basis_name)

        self._self_to_base_cache = {}
        self._base_to_self_cache = {}
        self.register_coercion(SetMorphism(Hom(base, self), self._base_to_self))
        base.register_coercion(SetMorphism(Hom(self, base), self._self_to_base))

    def construction(self):
        """
        Return a pair ``(F, R)``, where ``F`` is a
        :class:`SymmetricFunctionsFunctor` and `R` is a ring, such
        that ``F(R)`` returns ``self``.

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur')
            sage: s.construction()
            (SymmetricFunctionsFunctor[Schur], Rational Field)
        """
        return OrthotriangBasisFunctor(self), self.base_ring()

    def _base_to_self(self, x):
        """
        Coerce a symmetric function in base ``x`` into ``self``.

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``x`` -- an element of the basis `base` to which ``self`` is triangularly related

        OUTPUT: an element of ``self`` equivalent to ``x``

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: s._base_to_self(m([2,1]))
            -2*s[1, 1, 1] + s[2, 1]
        """
        return self._from_cache(x, self._base_cache, self._base_to_self_cache)

    def _self_to_base(self, x):
        """
        Coerce a symmetric function in ``self`` into base ``x``.

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``x`` -- an element of ``self`` as a basis of the ring of symmetric functions

        OUTPUT: the element ``x`` expressed in the basis to which ``self`` is triangularly related

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: s._self_to_base(s([2,1]))
            2*m[1, 1, 1] + m[2, 1]
        """
        return self._sf_base._from_cache(x, self._base_cache, self._self_to_base_cache)

    def _base_cache(self, n):
        """
        Compute the change of basis between ``self`` and base for the
        homogeneous component of size ``n``

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``n`` -- nonnegative integer

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: s._base_cache(2)
            sage: l = lambda c: [ (i[0],[j for j in sorted(i[1].items())]) for i in sorted(c.items())]
            sage: l(s._base_to_self_cache[2])
            [([1, 1], [([1, 1], 1)]), ([2], [([1, 1], -1), ([2], 1)])]
            sage: l(s._self_to_base_cache[2])
            [([1, 1], [([1, 1], 1)]), ([2], [([1, 1], 1), ([2], 1)])]
        """
        if n in self._self_to_base_cache:
            return
        self._self_to_base_cache[n] = {}

        self._gram_schmidt(n, self._sf_base, self._scalar,
                           self._self_to_base_cache,
                           leading_coeff=self._leading_coeff,
                           upper_triangular=True)
        self._invert_morphism(n, self.base_ring(), self._self_to_base_cache,
                              self._base_to_self_cache,
                              to_other_function=self._to_base)

    def _to_base(self, part):
        r"""
        Return a function which takes in a partition `\mu` and returns the
        coefficient of a partition in the expansion of ``self`` `(part)` in base.

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``part`` -- a partition

        .. NOTE::

           We assume that self._gram_schmidt has been called before
           self._to_base is called.

        OUTPUT:

        - a function which accepts a partition ``mu`` and returns the coefficients
          in the expansion of ``self(part)`` in the triangularly related basis.

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: m(s([2,1]))
            2*m[1, 1, 1] + m[2, 1]
            sage: f = s._to_base(Partition([2,1]))
            sage: [f(p) for p in Partitions(3)]
            [0, 1, 2]
        """
        f = lambda mu: self._self_to_base_cache[part].get(mu, 0)
        return f

    def product(self, left, right):
        """
        Return ``left`` * ``right`` by converting both to the base and then
        converting back to ``self``.

        INPUT:

        - ``self`` -- a basis determined by an orthotriangular definition
        - ``left``, ``right`` -- elements in ``self``

        OUTPUT: the expansion of the product of ``left`` and ``right`` in the basis ``self``

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: s([1])*s([2,1]) #indirect doctest
            s[2, 1, 1] + s[2, 2] + s[3, 1]
        """
        return self(self._sf_base(left) * self._sf_base(right))


from sage.combinat.sf.sfa import SymmetricFunctionsFunctor


class OrthotriangBasisFunctor(SymmetricFunctionsFunctor):
    """
    A constructor for algebras of symmetric functions constructed by
    orthogonality and triangularity.

    EXAMPLES::

        sage: from sage.combinat.sf.sfa import zee
        sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
        sage: Sym = SymmetricFunctions(QQ)
        sage: m = Sym.m()
        sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur')
        sage: s.construction()
        (SymmetricFunctionsFunctor[Schur], Rational Field)
    """
    def __init__(self, basis):
        r"""
        Initialize the functor.

        INPUT:

        - ``basis`` -- the basis of the symmetric function algebra

        TESTS::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur')
            sage: F, R = s.construction()
            sage: TestSuite(F).run()
        """
        self._basis_name = basis.basis_name()
        self._sf_base = basis._sf_base
        self._scalar = basis._scalar
        self._leading_coeff = basis._leading_coeff
        self._prefix = basis._prefix
        super().__init__(basis, self._basis_name)

    def _apply_functor(self, R):
        """
        Apply the functor to an object of ``self``'s domain.

        EXAMPLES::

            sage: from sage.combinat.sf.sfa import zee
            sage: from sage.combinat.sf.orthotriang import SymmetricFunctionAlgebra_orthotriang
            sage: Sym = SymmetricFunctions(ZZ)
            sage: m = Sym.m()
            sage: s = SymmetricFunctionAlgebra_orthotriang(Sym, m, zee, 's', 'Schur functions')
            sage: F, R = s.construction()  # indirect doctest
            sage: F(QQbar)
            Symmetric Functions over Algebraic Field in the Schur functions basis
        """
        from sage.combinat.sf.sf import SymmetricFunctions
        return self._basis(SymmetricFunctions(R), self._sf_base.change_ring(R),
                           self._scalar, self._prefix, self._basis_name,
                           self._leading_coeff)


# Backward compatibility for unpickling
from sage.misc.persist import register_unpickle_override

register_unpickle_override('sage.combinat.sf.orthotriang', 'SymmetricFunctionAlgebraElement_orthotriang', SymmetricFunctionAlgebra_orthotriang.Element)



# sage.doctest: needs sage.libs.flint
"""
Morphisms of Hecke modules

AUTHORS:

- William Stein
"""

# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from sage.misc import misc
from sage.modules.matrix_morphism import MatrixMorphism
from sage.categories.morphism import Morphism

# We also define other types of Hecke-module morphisms that aren't
# specified by a matrix.  E.g., Hecke operators, or maybe morphisms on
# modular abelian varieties (which are specified by matrices, but on
# integral homology). All morphisms derive from HeckeModuleMorphism.


class HeckeModuleMorphism(Morphism):
    r"""
    Abstract base class for morphisms of Hecke modules.
    """
    pass


class HeckeModuleMorphism_matrix(MatrixMorphism, HeckeModuleMorphism):
    """
    Morphisms of Hecke modules when the morphism is given by a matrix.

    Note that care is needed when composing morphisms, because morphisms in
    Sage act on the left, but their matrices act on the right (!). So if F: A
    -> B and G : B -> C are morphisms, the composition A -> C is G*F, but its
    matrix is F.matrix() * G.matrix().

    EXAMPLES::

        sage: A = ModularForms(1, 4)
        sage: B = ModularForms(1, 16)
        sage: C = ModularForms(1, 28)
        sage: F = A.Hom(B)(matrix(QQ,1,2,srange(1, 3)))
        sage: G = B.Hom(C)(matrix(QQ,2,3,srange(1, 7)))
        sage: G * F
        Hecke module morphism defined by the matrix
        [ 9 12 15]
        Domain: Modular Forms space of dimension 1 for Modular Group SL(2,Z) ...
        Codomain: Modular Forms space of dimension 3 for Modular Group SL(2,Z) ...
        sage: F * G
        Traceback (most recent call last):
        ...
        TypeError: Incompatible composition of morphisms: domain of left morphism must be codomain of right.
    """
    def __init__(self, parent, A, name='', side='left') -> None:
        """
        INPUT:

        -  ``parent`` -- ModularSymbolsHomspace

        - ``A`` -- matrix

        - ``name`` -- string (default: ``''``); name of the morphism
           (used for printing)

        - ``side`` -- string (default: ``'left'``)

        EXAMPLES::

            sage: M = ModularSymbols(6)
            sage: t = M.Hom(M)(matrix(QQ,3,3,srange(9)), name='spam'); t
            Hecke module morphism spam defined by the matrix
            [0 1 2]
            [3 4 5]
            [6 7 8]
            Domain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...
            Codomain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...
            sage: t == loads(dumps(t))
            True
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string")
        self.__name = name
        MatrixMorphism.__init__(self, parent, A, side)

    def name(self, new=None):
        r"""
        Return the name of this operator, or set it to a new name.

        EXAMPLES::

            sage: M = ModularSymbols(6)
            sage: t = M.Hom(M)(matrix(QQ,3,3,srange(9)), name='spam'); t
            Hecke module morphism spam defined by ...
            sage: t.name()
            'spam'
            sage: t.name("eggs"); t
            Hecke module morphism eggs defined by ...
        """
        if new is None:
            return self.__name
        self.__name = new

    def _repr_(self):
        r"""
        String representation of ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(6)
            sage: t = M.Hom(M)(matrix(QQ,3,3,srange(9))); t._repr_()
            'Hecke module morphism defined by the matrix\n[0 1 2]\n[3 4 5]\n[6 7 8]\nDomain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...\nCodomain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...'
            sage: t.name('spam'); t._repr_()
            'Hecke module morphism spam defined by the matrix\n[0 1 2]\n[3 4 5]\n[6 7 8]\nDomain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...\nCodomain: Modular Symbols space of dimension 3 for Gamma_0(6) of weight ...'
        """
        name = self.__name
        if name:
            name += ' '
        return "Hecke module morphism %sdefined by the matrix\n%r\nDomain: %s\nCodomain: %s" % (name, self.matrix(), misc.strunc(self.domain()), misc.strunc(self.codomain()))

# __mul__ method removed by David Loeffler 2009-04-14 as it is an exact duplicate of sage.modules.matrix_morphism.__mul__



from sage.libs.gmp.types cimport mpq_t
from sage.modules.free_module_element cimport FreeModuleElement
from sage.structure.parent cimport Parent


cdef class Vector_rational_dense(FreeModuleElement):
    cdef mpq_t* _entries
    cdef int _init(self, Py_ssize_t degree, Parent parent) except -1

    cdef inline Vector_rational_dense _new_c(self):
        cdef type t = type(self)
        cdef Vector_rational_dense x = <Vector_rational_dense>(t.__new__(t))
        x._init(self._degree, self._parent)
        return x



r"""
Group homomorphisms for groups with a GAP backend

EXAMPLES::

    sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
    sage: A = AbelianGroupGap([2, 4])
    sage: F.<a,b> = FreeGroup()
    sage: f = F.hom([g for g in A.gens()])
    sage: K = f.kernel()
    sage: K
    Group(<free, no generators known>)

AUTHORS:

- Simon Brandhorst (2018-02-08): initial version
- Sebastian Oehms  (2018-11-15): have this functionality work for permutation groups (:issue:`26750`)
  and implement :meth:`section` and :meth:`natural_map`
"""

# ****************************************************************************
#       Copyright (C) 2018 Simon Brandhorst <sbrandhorst@web.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.homset import HomsetWithBase
from sage.categories.morphism import Morphism
from sage.rings.integer_ring import ZZ
from sage.groups.libgap_wrapper import ParentLibGAP
from sage.libs.gap.element import GapElement
from sage.misc.latex import latex


class GroupMorphism_libgap(Morphism):
    r"""
    This wraps GAP group homomorphisms.

    Checking if the input defines a group homomorphism can be expensive
    if the group is large.

    INPUT:

    - ``homset`` -- the parent
    - ``gap_hom`` -- a :class:`sage.libs.gap.element.GapElement` consisting of
      a group homomorphism
    - ``check`` -- boolean (default: ``True``); check if the ``gap_hom`` is a group
      homomorphism (this can be expensive)

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: A = AbelianGroupGap([2, 4])
        sage: A.hom([g^2 for g in A.gens()])
        Group endomorphism of Abelian group with gap, generator orders (2, 4)

    Homomorphisms can be defined between different kinds of GAP groups::

        sage: G = MatrixGroup([Matrix(ZZ, 2, [0,1,1,0])])
        sage: f = A.hom([G.0, G(1)])
        sage: f
        Group morphism:
        From: Abelian group with gap, generator orders (2, 4)
        To:   Matrix group over Integer Ring with 1 generators (
        [0 1]
        [1 0]
        )
        sage: G.<a,b> = FreeGroup()
        sage: H = G / (G([1]), G([2])^3)
        sage: f = G.hom(H.gens())
        sage: f
        Group morphism:
          From: Free Group on generators {a, b}
          To:   Finitely presented group < a, b | a, b^3 >

    Homomorphisms can be defined between GAP groups and permutation groups::

        sage: S = Sp(4,3)
        sage: P = PSp(4,3)
        sage: pr = S.hom(P.gens())
        sage: E = copy(S.one().matrix())
        sage: E[3,0] = 2; e = S(E)
        sage: pr(e)
        (1,16,15)(3,22,18)(4,19,21)(6,34,24)(7,25,33)(9,40,27)(10,28,39)(12,37,30)(13,31,36)

    TESTS:

    Old tests inherited from MatrixGroupMorphisms::

        sage: F = GF(5); MS = MatrixSpace(F,2,2)
        sage: G = MatrixGroup([MS([1,1,0,1])])
        sage: H = MatrixGroup([MS([1,0,1,1])])
        sage: phi = G.hom(H.gens())
        sage: phi
        Group morphism:
        From: Matrix group over Finite Field of size 5 with 1 generators (
        [1 1]
        [0 1]
        )
        To:   Matrix group over Finite Field of size 5 with 1 generators (
        [1 0]
        [1 1]
        )
        sage: phi(MS([1,1,0,1]))
        [1 0]
        [1 1]
        sage: F = GF(7); MS = MatrixSpace(F,2,2)
        sage: F.multiplicative_generator()                                              # needs sage.libs.pari
        3
        sage: G = MatrixGroup([MS([3,0,0,1])])
        sage: a = G.gens()[0]^2
        sage: phi = G.hom([a])

    Check that :issue:`19406` is fixed::

        sage: G = GL(2, GF(3))
        sage: H = GL(3, GF(2))
        sage: mat1 = H([[-1,0,0],[0,0,-1],[0,-1,0]])
        sage: mat2 = H([[1,1,1],[0,0,-1],[-1,0,0]])
        sage: phi = G.hom([mat1, mat2])
        Traceback (most recent call last):
        ...
        ValueError: images do not define a group homomorphism

        sage: F = GF(5); MS = MatrixSpace(F,2,2)
        sage: G = MatrixGroup([MS([1,1,0,1])])
        sage: H = MatrixGroup([MS([1,0,1,1])])
        sage: phi = G.hom(H.gens())
        sage: phi.gap()
        CompositionMapping( [ (6,7,8,10,9)(11,13,14,12,15)(16,19,20,18,17)(21,25,22,24,23) ]
        -> [ [ [ Z(5)^0, 0*Z(5) ], [ Z(5)^0, Z(5)^0 ] ] ], <action isomorphism> )
        sage: type(_)
        <class 'sage.libs.gap.element.GapElement'>

        sage: F = GF(7); MS = MatrixSpace(F,2,2)
        sage: F.multiplicative_generator()                                              # needs sage.libs.pari
        3
        sage: G = MatrixGroup([MS([3,0,0,1])])
        sage: a = G.gens()[0]^2
        sage: phi = G.hom([a])
        sage: phi.kernel()
        Subgroup with 1 generators (
        [6 0]
        [0 1]
        ) of Matrix group over Finite Field of size 7 with 1 generators (
        [3 0]
        [0 1]
        )

        sage: F = GF(7); MS = MatrixSpace(F,2,2)
        sage: F.multiplicative_generator()                                              # needs sage.libs.pari
        3
        sage: G = MatrixGroup([MS([3,0,0,1])])
        sage: a = G.gens()[0]^2
        sage: phi = G.hom([a])
        sage: phi
        Group endomorphism of Matrix group over Finite Field of size 7 with 1 generators (
        [3 0]
        [0 1]
        )
        sage: g = G.gens()[0]
        sage: phi(g)
        [2 0]
        [0 1]
        sage: H = MatrixGroup([MS(a.list())])
        sage: H
        Matrix group over Finite Field of size 7 with 1 generators (
        [2 0]
        [0 1]
        )

    The following tests against :issue:`10659`::

        sage: phi(H)   # indirect doctest
        Subgroup with 1 generators (
        [4 0]
        [0 1]
        ) of Matrix group over Finite Field of size 7 with 1 generators (
        [3 0]
        [0 1]
        )
        sage: F = GF(5); MS = MatrixSpace(F,2,2)
        sage: g = MS([1,1,0,1])
        sage: G = MatrixGroup([g])
        sage: phi = G.hom(G.gens())
        sage: phi(G.0)
        [1 1]
        [0 1]
        sage: phi(G(g^2))
        [1 2]
        [0 1]

        sage: F = GF(5); MS = MatrixSpace(F,2,2)
        sage: gens = [MS([1,2,  -1,1]),MS([1,1,  0,1])]
        sage: G = MatrixGroup(gens)
        sage: phi = G.hom(G.gens())
        sage: phi(G.0)
        [1 2]
        [4 1]
        sage: phi(G.1)
        [1 1]
        [0 1]

    We check that :issue:`19780` is fixed::

        sage: G = groups.matrix.SO(3, 3)
        sage: H = groups.matrix.GL(3, 3)
        sage: phi = G.hom([H(x) for x in G.gens()])
        sage: phi(G.one()).parent()
        General Linear Group of degree 3 over Finite Field of size 3

        sage: # needs sage.symbolic
        sage: MS = MatrixSpace(SR, 2, 2)
        sage: G = MatrixGroup([MS(1), MS([1,2,3,4])])
        sage: G.Hom(G)
        Set of Morphisms from Matrix group over Symbolic Ring with 2 generators (
        [1 0]  [1 2]
        [0 1], [3 4]
        ) to Matrix group over Symbolic Ring with 2 generators (
        [1 0]  [1 2]
        [0 1], [3 4]
        ) in Category of groups

        sage: G = MatrixGroup([matrix(GF(5), [[1,3],[0,1]])])
        sage: H = MatrixGroup([matrix(GF(5), [[1,2],[0,1]])])
        sage: G.hom([H.gen(0)])
        Group morphism:
        From: Matrix group over Finite Field of size 5 with 1 generators (
        [1 3]
        [0 1]
        )
        To:   Matrix group over Finite Field of size 5 with 1 generators (
        [1 2]
        [0 1]
        )

        sage: G = GO(2,2,e=1)
        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: A = AbelianGroupGap([2])
        sage: G.hom([A.one(),A.gen(0)])
        Group morphism:
        From: General Orthogonal Group of degree 2 and form parameter 1 over Finite Field of size 2
        To:   Abelian group with gap, generator orders (2,)

    Check that :issue:`19407` is fixed::

        sage: G = GL(2, GF(2))
        sage: H = GL(3, ZZ)
        sage: Hom(G, H)
        Set of Morphisms from General Linear Group of degree 2
         over Finite Field of size 2
         to General Linear Group of degree 3 over Integer Ring in Category of groups
    """
    def __init__(self, homset, gap_hom, check=True):
        r"""
        Constructor method.

        TESTS::

            sage: G = GL(2, ZZ)
            sage: H = GL(2, GF(2))
            sage: P = Hom(G, H)
            sage: gen1 = [g.gap() for g in G.gens()]
            sage: gen2 = [H(g).gap() for g in G.gens()]
            sage: phi = G.gap().GroupHomomorphismByImagesNC(H,gen1, gen2)
            sage: phi = P.element_class(P,phi)
            sage: phi(G.gen(0))
            [0 1]
            [1 0]
        """
        if check:
            if not gap_hom.IsGroupHomomorphism():
                raise ValueError("not a group homomorphism")
            if homset.domain().gap() != gap_hom.Source():
                raise ValueError("domains do not agree")
            if homset.codomain().gap() != gap_hom.Range():
                raise ValueError("ranges do not agree")
        Morphism.__init__(self, homset)
        self._phi = gap_hom

    def __reduce__(self):
        r"""
        Implement pickling.

        We have to work around the fact that GAP does not provide pickling.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,2,5])
            sage: f = G.hom(G, G.gens())
            sage: f == loads(dumps(f))
            True
        """
        return self.parent(), (tuple(self(g) for g in self.domain().gens()),)

    def _repr_type(self):
        r"""
        Part of the implementation of :meth:`_repr_`.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2, 4])
            sage: A.hom([g^2 for g in A.gens()]) # indirect doctest
            Group endomorphism of Abelian group with gap, generator orders (2, 4)
        """
        return "Group"

    def gap(self):
        r"""
        Return the underlying LibGAP group homomorphism.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: f = A.hom([g^2 for g in A.gens()])
            sage: f.gap()
            [ f1, f2 ] -> [ <identity> of ..., f3 ]
        """
        return self._phi

    def _latex_(self):
        r"""
        Return the latex representation of ``self``.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: f = A.hom([g^2 for g in A.gens()])
            sage: latex(f)
            \text{\texttt{Abelian group with gap, generator orders }} \left(2, 4\right)
             \rightarrow \text{\texttt{Abelian group with gap, generator orders }} \left(2, 4\right)
        """
        return r"{} \rightarrow {}".format(latex(self.domain()), latex(self.codomain()))

    def kernel(self):
        r"""
        Return the kernel of ``self``.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A1 = AbelianGroupGap([6, 6])
            sage: A2 = AbelianGroupGap([3, 3])
            sage: f = A1.hom(A2.gens())
            sage: f.kernel()
            Subgroup of Abelian group with gap, generator orders (6, 6)
             generated by (f1*f2, f3*f4)
            sage: f.kernel().order()
            4
            sage: S = Sp(6,3)
            sage: P = PSp(6,3)
            sage: pr = Hom(S, P).natural_map()
            sage: pr.kernel()
            Subgroup with 1 generators (
            [2 0 0 0 0 0]
            [0 2 0 0 0 0]
            [0 0 2 0 0 0]
            [0 0 0 2 0 0]
            [0 0 0 0 2 0]
            [0 0 0 0 0 2]
            ) of Symplectic Group of degree 6 over Finite Field of size 3
        """
        dom = self.domain()
        return dom._subgroup_constructor(self.gap().Kernel())

    def pushforward(self, J, *args, **kwds):
        r"""
        The image of an element or a subgroup.

        INPUT:

        - ``J`` -- a subgroup or an element of the domain of ``self``

        OUTPUT: the image of ``J`` under ``self``

        .. NOTE::

            ``pushforward`` is the method that is used when a map is called
            on anything that is not an element of its domain. For historical
            reasons, we keep the alias ``image()`` for this method.

        EXAMPLES::

            sage: G.<a,b> = FreeGroup()
            sage: H = G / (G([1]), G([2])^3)
            sage: f = G.hom(H.gens())
            sage: S = G.subgroup([a.gap()])
            sage: f.pushforward(S)
            Group([ a ])
            sage: x = f.image(a)
            sage: x
            a
            sage: x.parent()
            Finitely presented group < a, b | a, b^3 >

            sage: # needs sage.rings.finite_rings
            sage: G = GU(3,2)
            sage: P = PGU(3,2)
            sage: pr = Hom(G, P).natural_map()
            sage: GS = G.subgroup([G.gen(0)])
            sage: pr.pushforward(GS)
            Subgroup generated by [(3,4,5)(10,18,14)(11,19,15)(12,20,16)(13,21,17)] of
            (The projective general unitary group of degree 3 over Finite Field of size 2)
        """
        dom = self.domain()
        codom = self.codomain()
        phi = self.gap()
        if isinstance(J, dom.Element) and J in dom:
            return self._call_(dom(J))
        from sage.groups.perm_gps.permgroup import PermutationGroup_generic
        if not isinstance(J, (ParentLibGAP, PermutationGroup_generic)):
            raise TypeError("J (={}) must be a libgap or permutation group".format(J))
        if dom.gap().IsSubgroup(J.gap()).sage():
            return codom._subgroup_constructor(phi.Image(J.gap()))

    image = pushforward

    def _call_(self, g):
        """
        Call syntax for morphisms.

        Some python code for wrapping GAP's ``Images`` function.

        INPUT:

        - ``g`` -- an element of the domain

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2, 4])
            sage: f = A.hom([g^2 for g in A.gens()])
            sage: a = A.gens()[1]
            sage: f(a)
            f3

        TESTS:

        The following tests we do fall back behind :issue:`10659`::

            sage: O = WeylGroup(['D',6])
            sage: r = prod(O.gens())
            sage: r_ = r^-1
            sage: f = O.hom([r*x*r_ for x in O.gens()])  # long time (19s on sage.math, 2011)
            sage: [f(x) for x in O.gens()]  # long time
            [
            [1 0 0 0 0 0]  [1 0 0 0 0 0]  [1 0 0 0 0 0]  [ 0  0  0  0 -1  0]
            [0 0 1 0 0 0]  [0 1 0 0 0 0]  [0 1 0 0 0 0]  [ 0  1  0  0  0  0]
            [0 1 0 0 0 0]  [0 0 0 1 0 0]  [0 0 1 0 0 0]  [ 0  0  1  0  0  0]
            [0 0 0 1 0 0]  [0 0 1 0 0 0]  [0 0 0 0 1 0]  [ 0  0  0  1  0  0]
            [0 0 0 0 1 0]  [0 0 0 0 1 0]  [0 0 0 1 0 0]  [-1  0  0  0  0  0]
            [0 0 0 0 0 1], [0 0 0 0 0 1], [0 0 0 0 0 1], [ 0  0  0  0  0  1],
            <BLANKLINE>
            [0 0 0 0 0 1]  [ 0  0  0  0  0 -1]
            [0 1 0 0 0 0]  [ 0  1  0  0  0  0]
            [0 0 1 0 0 0]  [ 0  0  1  0  0  0]
            [0 0 0 1 0 0]  [ 0  0  0  1  0  0]
            [0 0 0 0 1 0]  [ 0  0  0  0  1  0]
            [1 0 0 0 0 0], [-1  0  0  0  0  0]
            ]
            sage: f(O)  # long time
            Subgroup with 6 generators of Weyl Group of type ['D', 6] (as a matrix group acting on the ambient space)
            sage: f(O).gens()   # long time
            (
            [1 0 0 0 0 0]  [1 0 0 0 0 0]  [1 0 0 0 0 0]  [ 0  0  0  0 -1  0]
            [0 0 1 0 0 0]  [0 1 0 0 0 0]  [0 1 0 0 0 0]  [ 0  1  0  0  0  0]
            [0 1 0 0 0 0]  [0 0 0 1 0 0]  [0 0 1 0 0 0]  [ 0  0  1  0  0  0]
            [0 0 0 1 0 0]  [0 0 1 0 0 0]  [0 0 0 0 1 0]  [ 0  0  0  1  0  0]
            [0 0 0 0 1 0]  [0 0 0 0 1 0]  [0 0 0 1 0 0]  [-1  0  0  0  0  0]
            [0 0 0 0 0 1], [0 0 0 0 0 1], [0 0 0 0 0 1], [ 0  0  0  0  0  1],
            <BLANKLINE>
            [0 0 0 0 0 1]  [ 0  0  0  0  0 -1]
            [0 1 0 0 0 0]  [ 0  1  0  0  0  0]
            [0 0 1 0 0 0]  [ 0  0  1  0  0  0]
            [0 0 0 1 0 0]  [ 0  0  0  1  0  0]
            [0 0 0 0 1 0]  [ 0  0  0  0  1  0]
            [1 0 0 0 0 0], [-1  0  0  0  0  0]
            )
        """
        img_gap = self.gap().Image(g.gap())
        return self.codomain()(img_gap)

    def lift(self, h):
        r"""
        Return an element of the domain that maps to ``h``.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: f = A.hom([g^2 for g in A.gens()])
            sage: a = A.gens()[1]
            sage: f.lift(a^2)
            f2

        If the element is not in the image, we raise an error::

            sage: f.lift(a)
            Traceback (most recent call last):
            ...
            ValueError: f2 is not an element of the image of Group endomorphism
             of Abelian group with gap, generator orders (2, 4)
        """
        if h not in self.codomain():
            raise TypeError("h (={}) must be an element of the codomain".format(h))
        h = self.codomain()(h)
        phi = self.gap()
        if h.gap() not in phi.Image():
            raise ValueError("{} is not an element of the image of {}".format(h, self))
        return self.domain()(phi.PreImagesRepresentative(h.gap()))

    def preimage(self, S):
        r"""
        Return the preimage of the subgroup ``S``.

        INPUT:

        - ``S`` -- a subgroup of this group

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: B = AbelianGroupGap([4])
            sage: f = A.hom([B.one(), B.gen(0)^2])
            sage: S = B.subgroup([B.one()])
            sage: f.preimage(S) == f.kernel()
            True
            sage: S = Sp(4,3)
            sage: P = PSp(4,3)
            sage: pr = Hom(S, P).natural_map()
            sage: PS = P.subgroup([P.gen(0)])
            sage: pr.preimage(PS)
            Subgroup with 2 generators (
            [2 0 0 0]  [1 0 0 0]
            [0 2 0 0]  [0 2 0 0]
            [0 0 2 0]  [0 0 2 0]
            [0 0 0 2], [0 0 0 1]
            ) of Symplectic Group of degree 4 over Finite Field of size 3
        """
        phi = self.gap()
        from sage.groups.perm_gps.permgroup import PermutationGroup_generic
        if not isinstance(S, (ParentLibGAP, PermutationGroup_generic)):
            raise TypeError("%s must be a GAP or permutation group of %s" % (S, self))
        if not self.codomain().gap().IsSubgroup(S.gap()).sage():
            raise ValueError("%s must be a subgroup of %s" % (S, self))
        preimage = phi.PreImage(S.gap())
        return self.domain()._subgroup_constructor(preimage)

    def section(self):
        r"""
        Return a section map of ``self`` by use of :meth:`lift`.

        See :meth:`section` of :class:`sage.categories.map.Map`, as well.

        OUTPUT: an instance of :class:`sage.categories.morphism.SetMorphism`
        mapping an element of the codomain of ``self`` to one of its preimages

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: G = GU(3,2)
            sage: P = PGU(3,2)
            sage: pr = Hom(G, P).natural_map()
            sage: sect = pr.section()
            sage: sect(P.an_element())
            [a + 1     a     a]
            [    1     1     0]
            [    a     0     0]
        """
        from sage.categories.homset import Hom
        from sage.categories.sets_cat import Sets
        H = Hom(self.codomain(), self.domain(), category=Sets())
        return H(self.lift)


class GroupHomset_libgap(HomsetWithBase):
    r"""
    Homsets of groups with a libgap backend.

    Do not call this directly instead use :func:`Hom`.

    INPUT:

    - ``G`` -- a libgap group
    - ``H`` -- a libgap group
    - ``category`` -- a category

    OUTPUT: the homset of two libgap groups

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: A = AbelianGroupGap([2,4])
        sage: H = A.Hom(A)
        sage: H
        Set of Morphisms from Abelian group with gap, generator orders (2, 4)
         to Abelian group with gap, generator orders (2, 4)
         in Category of finite enumerated commutative groups
    """
    def __init__(self, G, H, category=None, check=True):
        r"""
        Return the homset of two libgap groups.

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: H = A.Hom(A)
            sage: TestSuite(H).run()
        """
        if check:
            from sage.groups.perm_gps.permgroup import PermutationGroup_generic
            if not isinstance(G, (ParentLibGAP, PermutationGroup_generic)):
                raise TypeError("G (={}) must be a ParentLibGAP or a permutation group".format(G))
            if not isinstance(H, (ParentLibGAP, PermutationGroup_generic)):
                raise TypeError("H (={}) must be a ParentLibGAP or a permutation group".format(H))
        HomsetWithBase.__init__(self, G, H, category, check=check, base=ZZ)

    Element = GroupMorphism_libgap

    def _element_constructor_(self, x, check=True, **options):
        r"""
        Handle conversions and coercions.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: H = A.Hom(A)
            sage: H([g^2 for g in A.gens()])
            Group endomorphism of Abelian group with gap, generator orders (2, 4)
            sage: H(2)
            Traceback (most recent call last):
            ...
            TypeError: unable to convert 2 to an element of ...

        A group homomorphism between a finitely presented group and a subgroup of a permutation group::

            sage: # needs sage.rings.finite_rings
            sage: PG = PGU(6,2)
            sage: g, h = PG.gens()
            sage: p1 = h^-3*(h^-1*g^-1)^2*h*g*h^2*g^-1*h^2*g*h^-5*g^-1
            sage: p2 = g*(g*h)^2*g*h^-4*(g*h)^2*(h^2*g*h^-2*g)^2*h^-2*g*h^-2*g^-1*h^-1*g*h*g*h^-1*g
            sage: p3 = h^-3*g^-1*h*g*h^4*g^-1*h^-1*g*h*(h^2*g^-1)^2*h^-4*g*h^2*g^-1*h^-7*g^-2*h^-2*g*h^-2*g^-1*h^-1*(g*h)^2*h^3
            sage: p4 = h*(h^3*g)^2*h*g*h^-1*g*h^2*g^-1*h^-2*g*h^4*g^-1*h^3*g*h^-2*g*h^-1*g^-1*h^2*g*h*g^-1*h^-2*g*h*g^-1*h^2*g*h^2*g^-1
            sage: p5 = h^2*g*h^2*g^-1*h*g*h^-1*g*h*g^-1*h^2*g*h^-2*g*h^2*g*h^-2*(h^-1*g)^2*h^4*(g*h^-1)^2*g^-1
            sage: UPG = PG.subgroup([p1, p2, p3, p4, p5], canonicalize=False)
            sage: B6 = BraidGroup(6)
            sage: reprB6 = B6.hom(UPG.gens())
            sage: b1, b2, b3, b4, b5 = B6.gens()
            sage: reprB6(b1*b2*b3*b4*b5) == p1*p2*p3*p4*p5
            True

        TESTS:

        The following input does not define a valid group homomorphism::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2])
            sage: G = MatrixGroup([Matrix(ZZ, 2, [0,1,-1,0])])
            sage: A.hom([G.gen(0)])
            Traceback (most recent call last):
            ...
            ValueError: images do not define a group homomorphism
        """
        if isinstance(x, (tuple, list)):
            # there should be a better way
            from sage.libs.gap.libgap import libgap
            dom = self.domain()
            codom = self.codomain()
            gens = dom.gap().GeneratorsOfGroup()
            imgs = [codom(g).gap() for g in x]
            if check:
                if not len(gens) == len(imgs):
                    raise ValueError("provide an image for each generator")
                phi = libgap.GroupHomomorphismByImages(dom.gap(), codom.gap(), gens, imgs)
                # if it is not a group homomorphism, then
                # self._phi is the gap boolean fail
                if phi.is_bool():     # check we did not fail
                    raise ValueError("images do not define a group homomorphism")
            else:
                ByImagesNC = libgap.function_factory("GroupHomomorphismByImagesNC")
                phi = ByImagesNC(dom.gap(), codom.gap(), gens, imgs)
            return self.element_class(self, phi, check=check, **options)
        if isinstance(x, GapElement):
            try:
                return self.element_class(self, x, check=True, **options)
            except ValueError:
                pass
        return super()._element_constructor_(x, check=check, **options)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        There is only one group homomorphism which is always defined:
        the trivial one.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,4])
            sage: G.<a,b> = FreeGroup()
            sage: A.Hom(G).an_element()
            Group morphism:
              From: Abelian group with gap, generator orders (2, 4)
              To:   Free Group on generators {a, b}
        """
        n = len(self.domain().gens())
        one = self.codomain().one()
        return self._element_constructor_((one,) * n, check=False)

    def natural_map(self):
        r"""
        This method from :class:`HomsetWithBase` is overloaded here for cases in which
        both groups have corresponding lists of generators.

        OUTPUT:

        An instance of the element class of ``self`` if there exists a group
        homomorphism mapping the generators of the domain of ``self`` to the
        according generators of the codomain. Otherwise, the method falls back
        to the default.

        EXAMPLES::

            sage: G = GL(3,2)
            sage: P = PGL(3,2)
            sage: nat = Hom(G, P).natural_map()
            sage: type(nat)
            <class 'sage.groups.libgap_morphism.GroupHomset_libgap_with_category.element_class'>
            sage: g1, g2 = G.gens()
            sage: nat(g1*g2)
            (1,2,4,5,7,3,6)
        """
        if len(self.domain().gens()) == len(self.codomain().gens()):
            dom_gap = self.domain().gap()
            codom_gap = self.codomain().gap()
            from sage.libs.gap.libgap import libgap
            phi = libgap.GroupHomomorphismByImages(dom_gap,
                                                   codom_gap,
                                                   dom_gap.GeneratorsOfGroup(),
                                                   codom_gap.GeneratorsOfGroup()
                                                   )
            if not phi.is_bool():     # phi is indeed a group homomorphism
                return self.element_class(self, phi)
        return super().natural_map()



from typing import Any

from sage.structure.element import FieldElement

class FunctionFieldElement(FieldElement):
    _x: Any
    _matrix: Any

    def __reduce__(self) -> Any:
        ...

    def _new_c(self) -> 'FunctionFieldElement':
        ...

    def __pari__(self) -> Any:
        ...

    def _latex_(self) -> str:
        ...

    def subs(self, in_dict: dict | None = None, **kwds: Any) -> 'FunctionFieldElement':
        ...

    def matrix(self, base: Any = None) -> Any:
        ...

    def trace(self) -> Any:
        ...

    def norm(self) -> Any:
        ...

    def degree(self) -> int:
        ...

    def characteristic_polynomial(self, *args: Any, **kwds: Any) -> Any:
        ...

    def minimal_polynomial(self, *args: Any, **kwds: Any) -> Any:
        ...

    def is_integral(self) -> bool:
        ...

    def differential(self) -> Any:
        ...

    def derivative(self) -> Any:
        ...

    def higher_derivative(self, i: int, separating_element: Any = None) -> Any:
        ...

    def divisor(self) -> Any:
        ...

    def divisor_of_zeros(self) -> Any:
        ...

    def divisor_of_poles(self) -> Any:
        ...

    def zeros(self) -> list[Any]:
        ...

    def poles(self) -> list[Any]:
        ...

    def valuation(self, place: Any) -> int:
        ...

    def evaluate(self, place: Any) -> Any:
        ...

    def is_nth_power(self, n: int) -> bool:
        ...

    def nth_root(self, n: int) -> 'FunctionFieldElement':
        ...



# sage.doctest: needs sage.combinat sage.modules
"""
Abreu-Nigro symmetric functions
"""
# ****************************************************************************
#       Copyright (C) 2025 Travis Scrimshaw <tcscrims at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method

from . import multiplicative


class SymmetricFunctionAlgebra_AbreuNigro(multiplicative.SymmetricFunctionAlgebra_multiplicative):
    r"""
    The Abreu-Nigro (symmetric function) basis.

    The Abreu-Nigro basis `\{\rho_{\lambda}(x; q)\}_{\lambda}` is the
    multiplicative basis defined by

    .. MATH::

        [n]_q h_n(x) = \sum_{i=1}^n h_{n-i}(x) \rho_i(x; q),
        \qquad\qquad
        \rho_{\lambda}(x; q) = \rho_{\lambda_1}(x; q) \rho_{\lambda_2}(x; q)
        \cdots \rho_{\lambda_{\ell}}(x; q).

    Here `[n]_q = 1 + q + \cdots + q^{n-1}` is a `q`-integer.
    An alternative definition is given by
    `\rho_n = q^{n-1} P_n(x; q^{-1})`, where `P_n(x; q)` is the
    Hall-Littlewood `P` function for the one-row partition `n`.

    INPUT:

    - ``Sym`` -- the ring of the symmetric functions
    - ``q`` -- the parameter `q`

    REFERENCES:

    - [AN2021]_
    - [AN2021II]_
    - [AN2023]_

    EXAMPLES:

    We verify the change of basis formula for the first few `n`::

        sage: q = ZZ['q'].fraction_field().gen()
        sage: Sym = SymmetricFunctions(q.parent())
        sage: an = Sym.abreu_nigro(q)
        sage: h = Sym.h()
        sage: from sage.combinat.q_analogues import q_int, q_factorial
        sage: all(q_int(n, q) * h[n] == sum(h[n-i] * an[i] for i in range(1,n+1))
        ....:     for n in range(1, 5))
        True

        sage: P = Sym.hall_littlewood(q).P()
        sage: all(h(P[n]).map_coefficients(lambda c: q^(n-1) * c(q=~q)) == h(an[n])
        ....:     for n in range(1, 6))
        True

    Next, we give the expansion in a few other bases::

        sage: p = Sym.p()
        sage: s = Sym.s()
        sage: m = Sym.m()
        sage: e = Sym.e()

        sage: p(an([1]))
        p[1]
        sage: m(an([1]))
        m[1]
        sage: e(an([1]))
        e[1]
        sage: h(an([1]))
        h[1]
        sage: s(an([1]))
        s[1]

        sage: p(an([2]))
        ((q-1)/2)*p[1, 1] + ((q+1)/2)*p[2]
        sage: m(an([2]))
        (q-1)*m[1, 1] + q*m[2]
        sage: e(an([2]))
        q*e[1, 1] + (-q-1)*e[2]
        sage: h(an([2]))
        -h[1, 1] + (q+1)*h[2]
        sage: s(an([2]))
        -s[1, 1] + q*s[2]

        sage: p(an([3]))
        ((q^2-2*q+1)/6)*p[1, 1, 1] + ((q^2-1)/2)*p[2, 1] + ((q^2+q+1)/3)*p[3]
        sage: m(an([3]))
        (q^2-2*q+1)*m[1, 1, 1] + (q^2-q)*m[2, 1] + q^2*m[3]
        sage: e(an([3]))
        q^2*e[1, 1, 1] + (-2*q^2-q)*e[2, 1] + (q^2+q+1)*e[3]
        sage: h(an([3]))
        h[1, 1, 1] + (-q-2)*h[2, 1] + (q^2+q+1)*h[3]
        sage: s(an([3]))
        s[1, 1, 1] - q*s[2, 1] + q^2*s[3]

    Some examples of conversions the other way::

        sage: q_int(3, q) * an(h[3])
        (1/(q+1))*an[1, 1, 1] + ((q+2)/(q+1))*an[2, 1] + an[3]
        sage: q_int(3, q) * an(e[3])
        (q^3/(q+1))*an[1, 1, 1] + ((-2*q^2-q)/(q+1))*an[2, 1] + an[3]
        sage: q_int(3, q) * an(m[2,1])
        ((-2*q^3+q^2+q)/(q+1))*an[1, 1, 1] + ((5*q^2+2*q-1)/(q+1))*an[2, 1] - 3*an[3]
        sage: q_int(3, q) * an(p[3])
        (q^2-2*q+1)*an[1, 1, 1] + (-3*q+3)*an[2, 1] + 3*an[3]

    We verify the determinant formulas of [AN2021II]_ Proposition 2.1, but
    correcting a parity issue with (3)::

        sage: def h_det(n):
        ....:     ret = matrix.zero(an, n)
        ....:     for i in range(n):
        ....:         if i != 0:
        ....:             ret[i,i-1] = -q_int(i)
        ....:         for j in range(i, n):
        ....:             ret[i,j] = an[j-i+1]
        ....:     return ret.det()
        sage: all(q_factorial(n, q) * h[n] == h(h_det(n)) for n in range(6))
        True
        sage: all(q_factorial(n, q) * an(h[n]) == h_det(n) for n in range(6))
        True

        sage: def rho_det(n):
        ....:     ret = matrix.zero(h, n)
        ....:     for i in range(n):
        ....:         if i == 0:
        ....:             for j in range(n):
        ....:                 ret[0,j] = q_int(j+1, q) * h[j+1]
        ....:         else:
        ....:             for j in range(i-1, n):
        ....:                 ret[i,j] = h[j-i+1]
        ....:     return ret.det()
        sage: all((-1)^(n+1) * an[n] == an(rho_det(n)) for n in range(1, 6))
        True
        sage: all((-1)^(n+1) * h(an[n]) == rho_det(n) for n in range(1, 6))
        True

    Antipodes::

        sage: an([1]).antipode()
        -an[1]
        sage: an([2]).antipode()
        (q-1)*an[1, 1] - an[2]
        sage: an([3]).antipode()
        (-q^2+2*q-1)*an[1, 1, 1] + (2*q-2)*an[2, 1] - an[3]

    For single row partitions, the antipode is given by the formula
    `S(\rho_n(x; q)) = -P_n(x; q)`, where `P_n`
    is the Hall-Littlewood P-function::

        sage: P = Sym.hall_littlewood(q).P()
        sage: all(P(an[n].antipode()) == -P[n] for n in range(1, 6))
        True
    """
    @staticmethod
    def __classcall_private__(cls, Sym, q='q'):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: Sym = SymmetricFunctions(q.parent())
            sage: Sym.abreu_nigro(q) is Sym.abreu_nigro('q')
            True
        """
        q = Sym.base_ring()(q)
        return super().__classcall__(cls, Sym, q)

    def __init__(self, Sym, q):
        r"""
        Initialize ``self``.

        TESTS::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: Sym = SymmetricFunctions(q.parent())
            sage: an = Sym.abreu_nigro(q)
            sage: TestSuite(an).run(skip=['_test_associativity', '_test_distributivity', '_test_prod'])
            sage: TestSuite(an).run(elements=[an[1,1]+an[2], an[1]+2*an[1,1]])
            sage: latex(an[2,1])
            \rho_{2,1}
        """
        self._q = q
        multiplicative.SymmetricFunctionAlgebra_multiplicative.__init__(self, Sym, "Abreu-Nigro", 'an')
        self._print_options['latex_prefix'] = "\\rho"

        self._h = Sym.h()
        self.register_coercion(self._h._module_morphism(self._h_to_an_on_basis, codomain=self))
        self._h.register_coercion(self._module_morphism(self._an_to_h_on_basis, codomain=self._h))

    @cached_method
    def _h_to_an_on_basis(self, lam):
        r"""
        Return the complete homogeneous symmetric function ``h[lam]``
        expanded in the Abreu-Nigro basis.

        INPUT:

        - ``lam`` -- a partition

        EXAMPLES::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: Sym = SymmetricFunctions(q.parent())
            sage: an = Sym.abreu_nigro(q)
            sage: h = Sym.homogeneous()
            sage: an._h_to_an_on_basis(Partition([]))
            an[]
            sage: from sage.combinat.q_analogues import q_factorial as qfact
            sage: qfact(4, q) * qfact(2, q) * an._h_to_an_on_basis(Partition([4,2,1]))
            an[1, 1, 1, 1, 1, 1, 1] + (q^2+2*q+4)*an[2, 1, 1, 1, 1, 1]
             + (2*q^2+3*q+4)*an[2, 2, 1, 1, 1] + (q^2+q+1)*an[2, 2, 2, 1]
             + (q^3+2*q^2+3*q+2)*an[3, 1, 1, 1, 1] + (q^3+2*q^2+3*q+2)*an[3, 2, 1, 1]
             + (q^3+2*q^2+2*q+1)*an[4, 1, 1, 1] + (q^3+2*q^2+2*q+1)*an[4, 2, 1]
            sage: h(an._h_to_an_on_basis(Partition([3,1]))) == h[3,1]
            True

            sage: all(an(h(an[n])) == an[n] for n in range(10))
            True
        """
        if not lam:
            return self.one()
        P = self._indices
        if len(lam) == 1:
            R = self.base_ring()
            q = self._q
            B = self.basis()
            n = lam[0]
            return (self.sum(self._h_to_an_on_basis(P([n-i])) * B[P([i])]
                             for i in range(1, n+1)) / R.sum(q**k for k in range(n)))
        # Multiply by the smallest part to minimize the number of products
        return self._h_to_an_on_basis(P(lam[:-1])) * self._h_to_an_on_basis(P([lam[-1]]))

    @cached_method
    def _an_to_h_on_basis(self, lam):
        r"""
        Return the Abreu-Nigro symmetric function ``an[lam]``  expanded in the
        complete homogeneous basis.

        INPUT:

        - ``lam`` -- a partition

        EXAMPLES::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: Sym = SymmetricFunctions(q.parent())
            sage: an = Sym.abreu_nigro(q)
            sage: h = Sym.homogeneous()
            sage: an._an_to_h_on_basis(Partition([]))
            h[]
            sage: an._an_to_h_on_basis(Partition([4,2,1]))
            h[1, 1, 1, 1, 1, 1, 1] + (-2*q-4)*h[2, 1, 1, 1, 1, 1]
             + (q^2+5*q+4)*h[2, 2, 1, 1, 1] + (-q^2-2*q-1)*h[2, 2, 2, 1]
             + (q^2+q+2)*h[3, 1, 1, 1, 1] + (-q^3-2*q^2-3*q-2)*h[3, 2, 1, 1]
             + (-q^3-q^2-q-1)*h[4, 1, 1, 1] + (q^4+2*q^3+2*q^2+2*q+1)*h[4, 2, 1]
            sage: an(an._an_to_h_on_basis(Partition([3,1]))) == an[3,1]
            True

            sage: all(h(an(h[n])) == h[n] for n in range(10))
            True
        """
        if not lam:
            return self._h.one()
        P = self._indices
        if len(lam) == 1:
            R = self.base_ring()
            q = self._q
            B = self._h.basis()
            n = lam[0]
            return (R.sum(q**k for k in range(n)) * self._h[n]
                    - self._h.sum(B[P([n-i])] * self._an_to_h_on_basis(P([i])) for i in range(1, n)))
        # Multiply by the smallest part to minimize the number of products
        return self._an_to_h_on_basis(P(lam[:-1])) * self._an_to_h_on_basis(P([lam[-1]]))

    def coproduct_on_generators(self, n):
        r"""
        Return the coproduct on the ``n``-th generator of ``self``.

        For any `n \geq 1`, we have

        .. MATH::

            \Delta(\rho_n) = \rho_0 \otimes \rho_n
            + (q-1) \sum_{k=1}^{n-1} \rho_k \otimes \rho_{n-k}
            + \rho_n \otimes \rho_0.

        INPUT:

        - ``n`` -- a nonnegative integer

        OUTPUT:

        an element of the tensor squared of the basis ``self``

        EXAMPLES::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: Sym = SymmetricFunctions(q.parent())
            sage: h = Sym.h()
            sage: an = Sym.abreu_nigro(q)
            sage: an.coproduct_on_generators(2)
            an[] # an[2] + (q-1)*an[1] # an[1] + an[2] # an[]
            sage: an[2].coproduct()
            an[] # an[2] + (q-1)*an[1] # an[1] + an[2] # an[]
            sage: an.coproduct(an[2])
            an[] # an[2] + (q-1)*an[1] # an[1] + an[2] # an[]
            sage: an.tensor_square()(h(an[5]).coproduct()) == an[5].coproduct()
            True
            sage: an[2,1].coproduct()
            an[] # an[2, 1] + (q-1)*an[1] # an[1, 1] + an[1] # an[2]
             + (q-1)*an[1, 1] # an[1] + an[2] # an[1] + an[2, 1] # an[]
            sage: an.tensor_square()(h(an[2,1]).coproduct())
            an[] # an[2, 1] + (q-1)*an[1] # an[1, 1] + an[1] # an[2]
             + (q-1)*an[1, 1] # an[1] + an[2] # an[1] + an[2, 1] # an[]
        """
        TS = self.tensor_square()
        if not n:
            return TS.one()
        P = self._indices
        one = self.base_ring().one()
        d = {(P([n]), P([])): one, (P([]), P([n])): one}
        coeff = self._q - one
        if coeff:
            for k in range(1, n):
                d[P([k]), P([n-k])] = coeff
        return TS.element_class(TS, d)



# sage.doctest: needs sage.libs.flint sage.graphs sage.modules
r"""
Some fast computations for finite posets using FLINT matrices
"""
# ****************************************************************************
#       Copyright (C) 2020 Frédéric Chapoton <chapoton@unistra.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from cysignals.signals cimport sig_check
from cysignals.memory cimport sig_malloc, sig_free

from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpz_mat cimport *
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.matrix_space import MatrixSpace
from sage.rings.integer_ring import ZZ
from sage.libs.flint.fmpz_poly_sage cimport Fmpz_poly


cpdef Fmpz_poly chain_poly(list positions):
    r"""
    Return the chain polynomial of a poset.

    INPUT:

    - ``positions`` -- a list of sets of integers describing the poset, as
      given by the lazy attribute ``_leq_storage`` of Hasse diagrams

    OUTPUT: a Flint polynomial in one variable over `\ZZ`.

    EXAMPLES::

        sage: from sage.combinat.posets.hasse_cython_flint import chain_poly
        sage: D = [{0, 1}, {1}]
        sage: chain_poly(D)
        3  1 2 1
        sage: P = posets.TamariLattice(5)
        sage: H = P._hasse_diagram
        sage: D = H._leq_storage
        sage: chain_poly(D)
        12  1 42 357 1385 3133 4635 4758 3468 1778 612 127 12
    """
    cdef Py_ssize_t n = len(positions)
    cdef Py_ssize_t i, j

    q = Fmpz_poly([0, 1])
    zero = Fmpz_poly(0)
    one = Fmpz_poly(1)

    cdef list chain_polys = [zero] * n

    # chain_polys[i] will be the generating function for the
    # chains with lowest vertex i (in the labelling of the
    # Hasse diagram).
    for i in range(n - 1, -1, -1):
        cpi = q
        for j in positions[i]:
            cpi += q * chain_polys[j]
        chain_polys[i] = cpi
    total = one
    for i in range(n):
        total += chain_polys[i]
    return total


cpdef Matrix_integer_dense moebius_matrix_fast(list positions):
    """
    Compute the Möbius matrix of a poset by a specific triangular inversion.

    INPUT:

    - ``positions`` -- a list of sets of integers describing the poset, as
      given by the lazy attribute ``_leq_storage`` of Hasse diagrams

    OUTPUT: a dense matrix

    EXAMPLES::

        sage: from sage.combinat.posets.hasse_cython_flint import moebius_matrix_fast
        sage: D = [{0,1},{1}]
        sage: moebius_matrix_fast(D)
        [ 1 -1]
        [ 0  1]
        sage: P = posets.TamariLattice(5)
        sage: H = P._hasse_diagram
        sage: D = H._leq_storage
        sage: moebius_matrix_fast(D)
        42 x 42 dense matrix over Integer Ring (...)
    """
    cdef Matrix_integer_dense A
    cdef Py_ssize_t n = len(positions)
    cdef Py_ssize_t i
    cdef int j, k
    A = Matrix_integer_dense.__new__(Matrix_integer_dense,
                                     MatrixSpace(ZZ, n, n), None, None, None)
    fmpz_mat_one(A._matrix)
    cdef Py_ssize_t *pos_lens = <Py_ssize_t*> sig_malloc(n*sizeof(Py_ssize_t))
    cdef int **pos_array = <int**> sig_malloc(n*sizeof(int*))
    cdef Py_ssize_t jind, kind
    for i in range(n):
        pos_lens[i] = len(positions[i])
        pos_array[i] = <int*> sig_malloc(pos_lens[i]*sizeof(int))
        for jind, j in enumerate(positions[i]):
            pos_array[i][jind] = j

    for i in range(n - 1, -1, -1):
        sig_check()
        for jind in range(pos_lens[i]):
            j = pos_array[i][jind]
            if j != i:
                for kind in range(pos_lens[j]):
                    k = pos_array[j][kind]
                    fmpz_sub(fmpz_mat_entry(A._matrix, i, k),
                             fmpz_mat_entry(A._matrix, i, k),
                             fmpz_mat_entry(A._matrix, j, k))
    for i in range(n):
        sig_free(pos_array[i])
    sig_free(pos_array)
    sig_free(pos_lens)
    return A


cpdef Matrix_integer_dense coxeter_matrix_fast(list positions):
    """
    Compute the Coxeter matrix of a poset by a specific algorithm.

    INPUT:

    - ``positions`` -- a list of sets of integers describing the poset, as
      given by the lazy attribute ``_leq_storage`` of Hasse diagrams

    OUTPUT: a dense matrix

    EXAMPLES::

        sage: from sage.combinat.posets.hasse_cython_flint import coxeter_matrix_fast
        sage: D = [{0,1},{1}]
        sage: coxeter_matrix_fast(D)
        [ 0 -1]
        [ 1 -1]
        sage: P = posets.TamariLattice(5)
        sage: H = P._hasse_diagram
        sage: D = H._leq_storage
        sage: coxeter_matrix_fast(D)
        42 x 42 dense matrix over Integer Ring (...)
    """
    cdef Matrix_integer_dense A
    cdef Py_ssize_t n = len(positions)
    cdef Py_ssize_t i
    cdef int j, k
    A = Matrix_integer_dense.__new__(Matrix_integer_dense,
                                     MatrixSpace(ZZ, n, n), None, None, None)
    fmpz_mat_one(A._matrix)
    cdef Py_ssize_t *pos_lens = <Py_ssize_t*> sig_malloc(n*sizeof(Py_ssize_t))
    cdef int **pos_array = <int**> sig_malloc(n*sizeof(int*))
    cdef Py_ssize_t jind, kind
    for i in range(n):
        pos_lens[i] = len(positions[i])
        pos_array[i] = <int*> sig_malloc(pos_lens[i]*sizeof(int))
        for jind, j in enumerate(positions[i]):
            pos_array[i][jind] = j

    for i in range(n - 1, -1, -1):
        sig_check()
        for jind in range(pos_lens[i]):
            j = pos_array[i][jind]
            if j != i:
                for kind in range(pos_lens[j]):
                    k = pos_array[j][kind]
                    fmpz_sub(fmpz_mat_entry(A._matrix, k, i),
                             fmpz_mat_entry(A._matrix, k, i),
                             fmpz_mat_entry(A._matrix, k, j))
    for i in range(n):
        sig_check()
        for jind in range(pos_lens[i]):
            j = pos_array[i][jind]
            if j != i:
                for k in range(n):
                    fmpz_add(fmpz_mat_entry(A._matrix, i, k),
                             fmpz_mat_entry(A._matrix, i, k),
                             fmpz_mat_entry(A._matrix, j, k))
    for i in range(n):
        sig_free(pos_array[i])
    sig_free(pos_array)
    sig_free(pos_lens)
    fmpz_mat_scalar_mul_si(A._matrix, A._matrix, -1)
    return A



# ******************************************************************************
#       Copyright (C) 2020 Reimundo Heluani <heluani@potuz.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************


from sage.misc.lazy_import import lazy_import

lazy_import('sage.algebras.lie_conformal_algebras.lie_conformal_algebra',
            'LieConformalAlgebra')
lazy_import('sage.algebras.lie_conformal_algebras',
            'examples', 'lie_conformal_algebras')
del lazy_import



r"""
Interface to Groebner Fan

AUTHOR:

- Anders Nedergaard Jensen: Write gfan C++ program, which implements
  algorithms many of which were invented by Jensen, Komei
  Fukuda, and Rekha Thomas.
- William Stein (2006-03-18): wrote gfan interface (first version)
- Marshall Hampton (2008-03-17): modified to use gfan-0.3, subprocess instead of os.popen2

TODO -- much functionality of gfan-0.3 is still not exposed::

   * at most 52 variables:

       - use gfan_substitute to make easier (?)
       MH: I think this is now irrelevant since gfan can accept the original ring variables

   * --symmetry is really useful
            - permutations are 0-based *not* cycle notation; a <---> 0
     output is broken up much more nicely.

   * -- can work in Z/pZ for p <= 32749

   * -- can compute individual GB's for lex and revlex (via buchberger)
"""

# *****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from subprocess import Popen, PIPE

from sage.features.gfan import GfanExecutable


class Gfan:
    """
    Interface to Anders Jensen's Groebner Fan program.
    """
    def __call__(self, input, cmd='', verbose=False):
        r"""
        Call Groebner Fan program with given input.

        INPUT:

        - ``input`` -- string; input
        - ``cmd`` -- string (default: ``''``); GFan command
        - ``verbose`` -- boolean (default: ``False``)

        EXAMPLES::

            sage: print(gfan('Q[x,y]{x^2-y-1,y^2-xy-2/3}', cmd='bases'))                # needs gfan
            Q[x,y]
            {{
            y^4+4/9-7/3*y^2-y^3,
            x+5/2*y+3/2*y^2-3/2*y^3}
            ,
            {
            x^2-1-y,
            x*y+2/3-y^2,
            y^3-5/3*y-y^2-2/3*x}
            ,
            {
            x^2-1-y,
            y^2-2/3-x*y}
            ,
            {
            x^4+1/3+x-2*x^2-x^3,
            y+1-x^2}
            }
        """
        if cmd:
            cmd = cmd.split(' ')
            cmd[0] = GfanExecutable(cmd[0]).absolute_filename()
        else:
            cmd = [GfanExecutable().absolute_filename()]

        if verbose:
            print("gfan command:\n%s" % cmd)
            print("gfan input:\n%s" % input)

        gfan_processes = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE,
                               encoding='latin-1')
        ans, err = gfan_processes.communicate(input=input)

        # sometimes, gfan outputs stuff to stderr even though
        # everything is fine ; we avoid interpreting this as an error
        if len(err) > 0 and not err.startswith('_application PolyhedralCone'):
            raise RuntimeError(err)

        return ans


# The instance
gfan = Gfan()
